from ...topos.graphs.trees import Tree
from ..instruments.instrument import Instrument
import pandas as pd


class ParameterTree(Tree):
    def __init__(self, root, children:tuple):
        super().__init__(root, children)
        for node in self.graph.nodes:
            self.graph.nodes[node].pop('label', None)
        self._meta['pfields'] = pd.Series([set()], index=[''])
        self._node_instruments = {}
        self._subtree_muted_pfields = {}
    
    def __getitem__(self, node):
        return ParameterNode(self, node)
    
    @property
    def pfields(self):
        return sorted(self._meta.loc['', 'pfields'])
    
    def _traverse_to_instrument_node(self, node):
        current = node
        while current is not None:
            if current in self._node_instruments:
                return current
            try:
                current = self.parent(current)
            except (IndexError, AttributeError, TypeError):
                try:
                    parents = list(self.predecessors(current))
                    current = parents[0] if parents else None
                except (IndexError, AttributeError):
                    current = None
        return None
    
    def set_pfields(self, node, **kwargs):
        self._meta.loc['', 'pfields'].update(kwargs.keys())
        
        for key, value in kwargs.items():
            self.graph.nodes[node][key] = value
        
        for descendant in self.descendants(node):
            descendant_data = self.graph.nodes[descendant]
            for key, value in kwargs.items():
                descendant_data[key] = value
    
    def set_instrument(self, node, instrument, exclude=None):        
        if not isinstance(instrument, Instrument):
            raise TypeError("Expected Instrument instance")
        
        if exclude is None:
            exclude = set()
        elif isinstance(exclude, str):
            exclude = {exclude}
        elif isinstance(exclude, (list, tuple)):
            exclude = set(exclude)
        elif not isinstance(exclude, set):
            exclude = set(exclude)
            
        instrument_pfields = set(instrument.keys())
        self._meta.loc['', 'pfields'].update(instrument_pfields)
        
        self._node_instruments[node] = instrument
        
        descendants = list(self.descendants(node))
        subtree_nodes = [node] + descendants
        
        existing_pfields = set()
        for n in subtree_nodes:
            existing_pfields.update(self.graph.nodes[n].keys())
        
        non_instrument_pfields = existing_pfields - instrument_pfields
        self._subtree_muted_pfields[node] = non_instrument_pfields
        
        for n in subtree_nodes:
            node_data = self.graph.nodes[n]
            for key in instrument.keys():
                if key in exclude:
                    node_data[key] = instrument[key]
                elif key == 'synth_name' or key not in node_data:
                    node_data[key] = instrument[key]
    
    def get_active_instrument(self, node):
        instrument_node = self._traverse_to_instrument_node(node)
        return self._node_instruments.get(instrument_node) if instrument_node else None
    
    def get_governing_subtree_node(self, node):
        return self._traverse_to_instrument_node(node)
    
    def get_active_pfields(self, node):
        active_instrument = self.get_active_instrument(node)
        if active_instrument is None:
            return list(self.items(node).keys())
        return list(active_instrument.keys())
        
    def get(self, node, key):
        return self.graph.nodes[node].get(key)
    
    def clear(self, node=None):
        if node is None:
            for n in self.graph.nodes:
                self.graph.nodes[n].clear()
        else:
            self.graph.nodes[node].clear()
            for descendant in self.descendants(node):
                self.graph.nodes[descendant].clear()
            
    def items(self, node):
        return dict(self.graph.nodes[node])

class ParameterNode:
    def __init__(self, tree, node):
        self._tree = tree
        self._node = node
        
    def __getitem__(self, key):
        if isinstance(key, str):
            return self._tree.get(self._node, key)
        raise TypeError("Key must be a string")
    
    def __setitem__(self, key, value):
        self._tree.set_pfields(self._node, **{key: value})
    
    def set_pfields(self, **kwargs):
        self._tree.set_pfields(self._node, **kwargs)
        
    def set_instrument(self, instrument, exclude=None):
        self._tree.set_instrument(self._node, instrument, exclude=exclude)
        
    def clear(self):
        self._tree.clear(self._node)
        
    def items(self):
        return self._tree.items(self._node)
    
    def active_items(self):
        all_items = self._tree.items(self._node)
        governing_subtree_node = self._tree.get_governing_subtree_node(self._node)
        
        if governing_subtree_node is None:
            return all_items
        
        muted_pfields = self._tree._subtree_muted_pfields.get(governing_subtree_node, set())
        return {k: v for k, v in all_items.items() if k not in muted_pfields}
    
    def __dict__(self):
        return self._tree.items(self._node)
        
    def __str__(self):
        return str(self.active_items())
    
    def __repr__(self):
        return repr(self.active_items())



# ------------------------------------------------------------------------------------
# Klotho/klotho/skora/skora.py
# ------------------------------------------------------------------------------------
'''
--------------------------------------------------------------------------------------
Notation and visualization tools.
--------------------------------------------------------------------------------------
'''
from klotho.thetos.parameters.instruments import PFIELDS
from klotho.utils.data_structures.dictionaries import SafeDict

import numpy as np
import pandas as pd
import regex as re
import json
import matplotlib.pyplot as plt
from matplotlib import cm

import os
from pathlib import Path

def set_score_path(synth_dir: str = 'tutorials/audiovisual', synth_name: str = 'Integrated'):
    current_path = Path(os.getcwd())
    if current_path.name != 'AlloPy':
        raise EnvironmentError("This script must be run from within the 'AlloPy/' directory.")

    while current_path.name != 'allolib_playground' and current_path.parent != current_path:
        current_path = current_path.parent

    if current_path.name != 'allolib_playground':
        raise FileNotFoundError("allolib_playground directory not found in the path hierarchy.")
    score_path = os.path.join(current_path, synth_dir, 'bin', f'{synth_name}-data')
    # return current_path / f'{synth_dir}/bin/{synth_name}-data'
    return score_path

def extract_pfields(filepath):
    with open(filepath, 'r') as file:
        cpp_contents = file.read()

    instruments = {}

    class_start_pattern = re.compile(r'class (\w+) :')
    param_pattern = re.compile(r'createInternalTriggerParameter\(\s*"(\w+)"\s*,\s*([\d.+-]+)')

    class_starts = [m.start() for m in class_start_pattern.finditer(cpp_contents)]
    class_starts.append(len(cpp_contents))

    for i in range(len(class_starts)-1):
        class_block = cpp_contents[class_starts[i]:class_starts[i+1]]
        class_name_match = class_start_pattern.search(class_block)
        if class_name_match:
            class_name = class_name_match.group(1)
            init_start_match = re.search(r'(virtual\s+)?void\s+init\(\)', class_block)
            if init_start_match:
                init_start = init_start_match.start()
                init_end = class_block.find('}', init_start)
                init_block = class_block[init_start:init_end]
                params = {"start": 0, "dur": 1, "synthName": class_name}  # standard pfields
                for param_match in param_pattern.finditer(init_block):
                    param_name, default_value_str = param_match.groups()
                    if param_name != "synthName":
                        default_value = float(default_value_str) if '.' in default_value_str else int(default_value_str)
                    else:
                        default_value = default_value_str
                    params[param_name] = default_value
                instruments[class_name] = params

    instruments_json = json.dumps(instruments, indent=4)
    target_directory = Path('./utils/instruments')
    target_directory.mkdir(parents=True, exist_ok=True)
    output_file_path = target_directory / (Path(filepath).stem + '.json')

    with open(output_file_path, 'w') as json_file:
        json_file.write(instruments_json)
        print(f'New JSON file created at: {output_file_path}')

def get_pfields(instrument_name, instruments_dir='utils/instruments', instrument_file='_instrument_classes.json'):
    # instruments_dir = './utils/instruments'
    base_name = os.path.splitext(os.path.basename(instrument_file))[0] + '.json'
    json_path = os.path.join(instruments_dir, base_name)

    # if not os.path.exists(json_path):
    #     extract_pfields(instrument_file)
    # extract_pfields(instrument_file)

    with open(json_path, 'r') as file:
        instruments_data = json.load(file)

    instrument_pfields = instruments_data.get(instrument_name, {})
    return SafeDict(instrument_pfields)

def make_score_df(pfields=('start', 'dur', 'synthName', 'amplitude', 'frequency')):
  '''
  Creates a DataFrame with the given pfields as columns.
  '''
  return pd.DataFrame(columns=pfields)

def synthSeq_to_df(filepath):
  '''
  Parses a .synthSequence file and returns the score as a Pandas DataFrame
  '''
  with open(filepath, 'r') as f:
      lines = f.readlines()

  params = []
  param_collect = False

  data = []
  # synths = set()
  for line in lines:
      if line.startswith('@'):
          line = line[2:]  # Remove '@ '
          components = line.split()
          data.append(components)
          # synths.add(components[2])
      elif line.startswith('#'):
        # synth parameters
        pattern = re.compile(r'^#\s+([a-zA-Z0-9\s]+(?:\s+[a-zA-Z0-9\s]+)*)\s*$') # pattern: `# {synthName}`
        match = pattern.match(line)
        if match:
          synth_name = components[2]
          params.extend(match.group(1).split()[1:])

  if not params:  # Default parameter names
      params = [f'synth_param_{str(i).zfill(2)}' for i in range(100)]

  df = pd.DataFrame(data)
  df.columns = ['start', 'dur', 'synthName'] + params[:df.shape[1]-3]

  return df

def df_to_synthSeq(df, filepath):
  '''
  Converts a DataFrame to a .synthSequence file
  '''
  with open(filepath, 'w') as f:
    for index, row in df.iterrows():
      f.write('@ ' + ' '.join(map(str, row.values)) + '\n')

def notelist_to_synthSeq(notelist, filepath):
  '''
  Converts a list of dictionaries to a .synthSequence file
  '''
  with open(filepath, 'w') as f:
    for row in notelist:
      f.write('@ ' + ' '.join(map(str, row.values())) + '\n')

def make_notelist(pfields: dict = {}, loop_param: str = 'max'):
  
  # if `pfields` is empty or incomplete, use default values
  if not pfields:
    pfields = getattr(PFIELDS, 'SineEnv', None).value.copy()
  elif 'synthName' not in pfields.keys():
    pfields['synthName'] = ['SineEnv']
  
  if loop_param == 'max':
    seq_len = max(len(value) if isinstance(value, list) else 1 for value in pfields.values())
  elif loop_param == 'min':
    seq_len = min(len(value) if isinstance(value, list) else 1 for value in pfields.values())
  elif loop_param in pfields.keys():
    seq_len = len(pfields[loop_param]) if isinstance(pfields[loop_param], list) else 1
  else:
    seq_len = 8

  pfields['start'] = 0.167 if 'start' not in pfields.keys() else pfields['start']
  pfields['dur']   = [1.0] if 'dur' not in pfields.keys() else pfields['dur'] if isinstance(pfields['dur'], list) else [pfields['dur']]
  pfields['dc']    = [1.0] if 'dc' not in pfields.keys() else pfields['dc'] if isinstance(pfields['dc'], list) else [pfields['dc']]

  note_list = []
  if not isinstance(pfields['synthName'], list):
    pfields['synthName'] = [pfields['synthName']]
  for i_syn, synthName in enumerate(pfields['synthName']):
    start = pfields['start'][0] if isinstance(pfields['start'], list) else pfields['start']
    for i in range(seq_len):
      new_row = getattr(PFIELDS, synthName, None).value.copy()
      new_row['start'] = start
      for key in pfields.keys():
        pfield = pfields[key] if isinstance(pfields[key], list) else [pfields[key]]
        plen = len(pfield)
        pidx = i % len(pfields[key]) if isinstance(pfields[key], list) else 0

        if key in ['start', 'synthName']:  # ignore these keys
          continue

        if key not in new_row.keys():  # check for name variations
          if key == 'amplitude':
            new_row['amp'] = pfields[key][pidx]
          if key == 'amp':
            new_row['amplitude'] = pfields[key][pidx]          
          if key == 'frequency':
            new_row['freq'] = pfields[key][pidx]
          if key == 'freq':
            new_row['frequency'] = pfields[key][pidx]
          continue
        
        # Get the pfield value
        if key == 'dur':
          new_row[key] = pfield[i % plen] * pfields['dc'][i % len(pfields['dc'])]  # apply duty cycle to duration
        else:
          new_row[key] = pfield[i % plen]  # set pfield value
      
      if new_row['dur'] > 0:                                       # negative durations mean rest (skip)
        note_list.append(new_row)                                  # append new row to the notelist
      if isinstance(pfields['start'], list):                       # if start is a list,
        start = pfields['start'][(i + 1) % len(pfields['start'])]  # get the next start time
      else:                                                        # otherwise,
        start += abs(pfields['dur'][i % len(pfields['dur'])])      # increment start time by the current duration

  return note_list

def play(pfields: dict = {}, loop_param: str = 'max', filename: str = 'play.synthSequence', inst: str = 'Integrated'):
  notelist = make_notelist(pfields=pfields, loop_param=loop_param)
  filepath = set_score_path(inst)
  notelist_to_synthSeq(notelist, os.path.join(filepath, filename))
  print(f'created "{filename}" in {filepath}...\n')

def make_row(rows_list: list, new_row: dict):
  '''
  
  '''
  rows_list.append(new_row)

def concat_rows(df, rows_list):
  '''
  Concatenate a list of rows to a DataFrame.
  
  Args:
  df (pd.DataFrame): The DataFrame to which to append the rows.
  rows_list (list): A list of rows to append to the DataFrame.
  
  Returns:
  pd.DataFrame: A new DataFrame with the rows appended.
  '''
  return pd.concat(
    [                 
      pd.DataFrame([row], 
                   columns=df.columns) for row in rows_list
    ],
    ignore_index=True)

def merge_parts_dfs(parts_list: list) -> pd.DataFrame:
  '''
  Merge a list of DataFrames into a single DataFrame.
  
  Args:
  parts_list (list): A list of DataFrames to merge.
  
  Returns:
  pd.DataFrame: A single DataFrame containing all the rows from the input DataFrames.
  '''
  return pd.concat(parts_list, ignore_index=True)

def get_score_duration(df: pd.DataFrame) -> float:
  '''
  Calculate the total duration of a `.synthSequence` file represented as a DataFrame.

  The duration is computed by finding the latest time point at which an event ends,
  which is the sum of its start time and duration.

  Args:
  df (pd.DataFrame): A DataFrame with at least 'start' and 'dur' columns representing the start time and duration of events.

  Returns:
  float: The total duration of the sequence.
  '''
  duration = 0.0
  for _, event in df.iterrows():
    d = float(event['start']) + float(event['dur'])
    if d > duration:
      duration = d
  return duration

def analyze_score(df):
  '''
  for each parameter in the dataframe.
  '''
  if not isinstance(df, pd.DataFrame):
    raise ValueError("Input must be a Pandas DataFrame.")

  # remove 'synthName' column
  param_cols = df.columns[:2].append(df.columns[3:])

  stats = {
      'min': [],
      'max': [],
      'mean': [],
      # etc...
      # continue to add more
  }

  for col in param_cols:
    stats['min'].append(df[col].astype(float).min())
    stats['max'].append(df[col].astype(float).max())
    stats['mean'].append(df[col].astype(float).mean())

  stats_df = pd.DataFrame(stats, index=param_cols)

  return stats_df

def plot_dataframe(df, column_name):
  fig, ax = plt.subplots(figsize=(16, 8))
  ax.set_facecolor('black')
  fig.patch.set_facecolor('black')

  df[column_name] = df[column_name].astype(float)
  # norm = plt.Normalize(vmin=np.log2(df[column_name].min()), vmax=np.log2(df[column_name].max()))
  # colors = cm.jet(norm(np.log2(df[column_name].values)))
  normed_freqs = np.log2(df[column_name].values) % 1
  colors = cm.jet(normed_freqs)

  for i, (idx, row) in enumerate(df.iterrows()):
    start_time = float(row['start'])
    end_time = start_time + float(row['dur'])# + float(row['releaseTime'])
    freq_val = row[column_name]

    ax.plot([start_time, end_time], [freq_val, freq_val], c=colors[i], linewidth=2)

  ax.set_yscale('log')
  min_freq = int(df[column_name].min())
  max_freq = int(df[column_name].max())

  ax.set_yticks([2**x for x in range(int(np.log2(min_freq)), int(np.log2(max_freq))+1)])
  ax.get_yaxis().set_major_formatter(plt.ScalarFormatter())

  ax.set_title(f'Plot of {column_name} over time (seconds)', color='white')
  ax.set_xlabel('time (seconds)', color='white')
  ax.set_ylabel(column_name + ' (Hz)', color='white')
  ax.tick_params(colors='white')
  ax.grid(True, which="both", ls="--", c='gray')



from .cps import CombinationProductSet
from .nkany import Hexany, Dekany, Pentadekany, Eikosany, Hebdomekontany

__all__ = [
    'CombinationProductSet',
    'Hexany',
    'Dekany',
    'Pentadekany',
    'Eikosany',
    'Hebdomekontany',
]


import numpy as np
from itertools import combinations
import pandas as pd
import math
from fractions import Fraction
from functools import cached_property
from typing import List, Tuple, Set, Dict, Any, Union, Literal
import networkx as nx
import sympy as sp
from scipy.spatial.distance import pdist, squareform


__all__ = [
    'Operations',
    'Sieve',
    'CombinationSet',
    'PartitionSet',
    'GenCol',
]

# ------------------------------------------------------------------------------
# Set Operations
# --------------

class Operations:
    """
    Static methods for mathematical set operations.

    This class provides a collection of static methods for performing
    common set operations including union, intersection, difference,
    and specialized musical set operations like transposition and inversion.

    Methods
    -------
    union(set1, set2)
        Return the union of two sets.
    intersect(set1, set2)
        Return the intersection of two sets.
    diff(set1, set2)
        Return the difference of two sets.
    symm_diff(set1, set2)
        Return the symmetric difference of two sets.
    is_subset(subset, superset)
        Check if the first set is a subset of the second set.
    is_superset(superset, subset)
        Check if the first set is a superset of the second set.
    invert(set1, axis, modulus)
        Invert a set around a given axis using modular arithmetic.
    transpose(set1, transposition_interval, modulus)
        Transpose a set by a given interval using modular arithmetic.
    complement(S, modulus)
        Return the complement of a set within a given modulus.
    congruent(S, modulus, residue)
        Return elements congruent to a residue modulo the given modulus.
    intervals(S)
        Calculate intervals between successive numbers in a sorted sequence.
    interval_vector(set1, modulus)
        Compute the interval vector of a set of pitches.

    Examples
    --------
    Basic set operations:
    
    >>> set1 = {0, 1, 4, 7}
    >>> set2 = {0, 2, 5, 9}
    >>> Operations.union(set1, set2)
    {0, 1, 2, 4, 5, 7, 9}
    
    Musical transformations:
    
    >>> chord = {0, 4, 7}  # C major triad
    >>> Operations.transpose(chord, 5, 12)  # Transpose up a fourth
    {5, 9, 0}
    """
    @staticmethod
    def union(set1: set, set2: set) -> set:        
        """
        Return the union of two sets.

        Parameters
        ----------
        set1 : set
            The first set.
        set2 : set
            The second set.

        Returns
        -------
        set
            A new set containing all unique elements from both sets.
        """
        return set1 | set2

    @staticmethod
    def intersect(set1: set, set2: set) -> set:
        """
        Return the intersection of two sets.

        Parameters
        ----------
        set1 : set
            The first set.
        set2 : set
            The second set.

        Returns
        -------
        set
            A new set containing only elements present in both sets.
        """
        return set1 & set2

    @staticmethod
    def diff(set1: set, set2: set) -> set:
        """
        Return the difference of two sets.

        Parameters
        ----------
        set1 : set
            The first set.
        set2 : set
            The second set.

        Returns
        -------
        set
            A new set containing elements in set1 but not in set2.
        """
        return set1 - set2

    @staticmethod
    def symm_diff(set1: set, set2: set) -> set:
        """
        Return the symmetric difference of two sets.

        Parameters
        ----------
        set1 : set
            The first set.
        set2 : set
            The second set.

        Returns
        -------
        set
            A new set containing elements in either set1 or set2 but not both.
        """
        return set1 ^ set2

    @staticmethod
    def is_subset(subset: set, superset: set) -> bool:
        """
        Check if the first set is a subset of the second set.

        Parameters
        ----------
        subset : set
            The potential subset.
        superset : set
            The potential superset.

        Returns
        -------
        bool
            True if subset is a subset of superset, False otherwise.
        """
        return subset <= superset

    @staticmethod
    def is_superset(superset: set, subset: set) -> bool:
        """
        Check if the first set is a superset of the second set.

        Parameters
        ----------
        superset : set
            The potential superset.
        subset : set
            The potential subset.

        Returns
        -------
        bool
            True if superset is a superset of subset, False otherwise.
        """
        return superset >= subset

    @staticmethod
    def invert(set1: set, axis: int = 0, modulus: int = 12) -> set:
        """
        Invert a set around a given axis using modular arithmetic.

        Parameters
        ----------
        set1 : set
            The set to invert.
        axis : int, optional
            The axis of inversion (default is 0).
        modulus : int, optional
            The modulus for the arithmetic (default is 12).

        Returns
        -------
        set
            The inverted set.

        Examples
        --------
        >>> chord = {0, 4, 7}  # C major triad
        >>> Operations.invert(chord, axis=0, modulus=12)
        {0, 8, 5}
        """
        return {(axis * 2 - pitch) % modulus for pitch in set1}

    @staticmethod
    def transpose(set1: set, transposition_interval: int, modulus: int = 12) -> set:
        """
        Transpose a set by a given interval using modular arithmetic.

        Parameters
        ----------
        set1 : set
            The set to transpose.
        transposition_interval : int
            The interval to transpose by.
        modulus : int, optional
            The modulus for the arithmetic (default is 12).

        Returns
        -------
        set
            The transposed set.

        Examples
        --------
        >>> chord = {0, 4, 7}  # C major triad
        >>> Operations.transpose(chord, 7, 12)  # Transpose up a fifth
        {7, 11, 2}
        """
        return {(pitch + transposition_interval) % modulus for pitch in set1}

    @staticmethod
    def complement(S: set, modulus: int = 12) -> set:
        """
        Return the complement of a set within a given modulus.

        Parameters
        ----------
        S : set
            The set to complement.
        modulus : int, optional
            The modulus defining the universal set (default is 12).

        Returns
        -------
        set
            The complement set containing all elements from 0 to modulus-1 not in S.

        Examples
        --------
        >>> pentatonic = {0, 2, 4, 7, 9}
        >>> Operations.complement(pentatonic, 12)
        {1, 3, 5, 6, 8, 10, 11}
        """
        return {s for s in range(modulus) if s not in S}

    @staticmethod
    def congruent(S: set, modulus: int, residue: int) -> set:
        """
        Return elements congruent to a residue modulo the given modulus.

        Parameters
        ----------
        S : set
            The input set.
        modulus : int
            The modulus for congruence testing.
        residue : int
            The residue value to match.

        Returns
        -------
        set
            Elements from S that are congruent to residue modulo modulus.

        Examples
        --------
        >>> numbers = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}
        >>> Operations.congruent(numbers, 3, 1)  # Numbers ≡ 1 (mod 3)
        {1, 4, 7}
        """
        return {s for s in S if s % modulus == residue}

    @staticmethod
    def intervals(S: set) -> set:
        """
        Calculate intervals between successive numbers in a sorted sequence.

        Parameters
        ----------
        S : set
            A set of numbers.

        Returns
        -------
        set
            The intervals between consecutive elements when sorted.

        Examples
        --------
        >>> scale = {0, 2, 4, 5, 7, 9, 11}  # Major scale
        >>> Operations.intervals(scale)
        {1, 2}
        """
        S = sorted(S)
        return set(np.diff(S))

    @staticmethod
    def interval_vector(set1: set, modulus: int = 12) -> np.ndarray:        
        """
        Compute the interval vector of a set of pitches.

        The interval vector represents the number of occurrences of each interval 
        between pitches in a set. Intervals larger than half the modulus are 
        inverted to their complements.

        Parameters
        ----------
        set1 : set
            A set of integers representing pitches.
        modulus : int, optional
            The modulus for interval calculations (default is 12).

        Returns
        -------
        numpy.ndarray
            An array representing the interval vector with length modulus//2.

        Examples
        --------
        >>> chord = {0, 4, 7}  # C major triad
        >>> Operations.interval_vector(chord, 12)
        array([0, 0, 1, 1, 1, 0])
        """
        pitches = sorted(set1)
        intervals = np.zeros(modulus // 2, dtype=int)

        for pitch1, pitch2 in combinations(pitches, 2):
            interval = abs(pitch2 - pitch1)
            interval = min(interval, modulus - interval)
            intervals[interval - 1] += 1

        return intervals


# ------------------------------------------------------------------------------
# Sieves
# ------

class Sieve:
    """
    Xenakis-style sieve for generating sets through modular arithmetic operations.

    A Sieve implements Iannis Xenakis's concept of sieves, which are mathematical
    structures based on modular arithmetic used in algorithmic composition. 
    A sieve can represent a single modular constraint or complex combinations 
    of multiple constraints using logical operations.

    In Xenakis's notation, a basic sieve is written as (m,r) where m is the 
    modulus and r is the residue. Complex sieves combine these using union (∪), 
    intersection (∩), and complement operations.

    Parameters
    ----------
    modulus : int, optional
        The step size of the arithmetic progression (default is 1).
    residue : int, optional
        The starting value of the progression (default is 0).
    N : int, optional
        The upper bound for generated values (default is 255).

    Attributes
    ----------
    S : set
        The generated sieve values.
    N : int
        The upper bound for generated values.
    period : int
        The modulus (step size) of the sieve.
    r : int
        The residue (starting value) of the sieve.
    congr : set
        Values congruent to the residue modulo the period.
    compl : set
        The complement of the sieve within [0, N].

    Examples
    --------
    Create a basic sieve (3,1) - multiples of 3 starting from 1:
    
    >>> sieve = Sieve(modulus=3, residue=1, N=10)
    >>> print(sorted(sieve.S))
    [1, 4, 7, 10]
    
    See Also
    --------
    Operations.union : Combine sieves using union operation
    Operations.intersect : Combine sieves using intersection operation
    
    Notes
    -----
    This implementation provides the foundation for Xenakis sieve theory.
    For complex sieve expressions, use the Operations class methods to
    combine multiple Sieve instances with logical operations.
    """
    def __init__(self, modulus: int = 1, residue: int = 0, N: int = 255):
        """
        Initialize a Sieve.

        Parameters
        ----------
        modulus : int, optional
            The step size of the arithmetic progression (default is 1).
        residue : int, optional
            The starting value of the progression (default is 0).
        N : int, optional
            The upper bound for generated values (default is 255).
        """
        self.__S = set(np.arange(residue, N + 1, modulus))
        self.__N = N
        self.__modulus = modulus
        self._residue = residue
    
    @property
    def S(self):
        """
        The generated sieve values.
        
        Returns
        -------
        set
            The set of integers generated by the sieve.
        """
        return self.__S
    
    @property
    def N(self):
        """
        The upper bound for generated values.
        
        Returns
        -------
        int
            The maximum value in the sieve range.
        """
        return self.__N
    
    @property
    def period(self):
        """
        The modulus (step size) of the sieve.
        
        Returns
        -------
        int
            The step size of the arithmetic progression.
        """
        return self.__modulus
    
    @property
    def r(self):
        """
        The residue (starting value) of the sieve.
        
        Returns
        -------
        int
            The starting value of the arithmetic progression.
        """
        return self._residue

    @property
    def congr(self):
        """
        Values congruent to the residue modulo the period.
        
        Returns
        -------
        set
            Elements in S that are congruent to residue mod period.
        """
        return Operations.congruent(self.__S, self.__modulus, self._residue)
    
    @property
    def compl(self):
        """
        The complement of the sieve within [0, N].
        
        Returns
        -------
        set
            All integers from 0 to N that are not in the sieve.
        """
        return Operations.complement(self.__S, self.__N)
    
    @N.setter
    def N(self, N: int):
        """
        Set the upper bound and regenerate the sieve.
        
        Parameters
        ----------
        N : int
            The new upper bound for the sieve.
        """
        self.__N = N
        self.__S = set(np.arange(self._residue, N + 1, self.__modulus))
    
    def __str__(self) -> str:
        """
        String representation of the Sieve.
        
        Returns
        -------
        str
            A formatted string showing the sieve parameters and values.
        """
        if len(self.__S) > 10:
            sieve = f'{list(self.__S)[:5]} ... {list(self.__S)[-1]}'
        else:
            sieve = list(self.__S)
        return (
            f'Period:  {self.__modulus}\n'
            f'Residue: {self._residue}\n'
            f'N:       {self.__N}\n'
            f'Sieve:   {sieve}\n'
        )

    def __repr__(self) -> str:        
        """
        String representation of the Sieve.
        
        Returns
        -------
        str
            A formatted string showing the sieve parameters and values.
        """
        return self.__str__()
    
    def __or__(self, other: 'Sieve') -> set:
        """
        Union operation between two sieves (Xenakis: A ∪ B).
        
        Parameters
        ----------
        other : Sieve
            The other sieve to combine with.
        
        Returns
        -------
        set
            The union of both sieve sets.
        
        Examples
        --------
        >>> sieve1 = Sieve(3, 1, 10)  # (3,1)
        >>> sieve2 = Sieve(5, 2, 10)  # (5,2)
        >>> combined = sieve1 | sieve2
        >>> print(sorted(combined))
        [1, 2, 4, 7, 10]
        """
        return Operations.union(self.S, other.S)
    
    def __and__(self, other: 'Sieve') -> set:
        """
        Intersection operation between two sieves (Xenakis: A ∩ B).
        
        Parameters
        ----------
        other : Sieve
            The other sieve to intersect with.
        
        Returns
        -------
        set
            The intersection of both sieve sets.
        
        Examples
        --------
        >>> sieve1 = Sieve(2, 0, 20)  # Even numbers
        >>> sieve2 = Sieve(3, 0, 20)  # Multiples of 3
        >>> intersection = sieve1 & sieve2
        >>> print(sorted(intersection))
        [0, 6, 12, 18]
        """
        return Operations.intersect(self.S, other.S)
    
    def __sub__(self, other: 'Sieve') -> set:
        """
        Difference operation between two sieves (Xenakis: A - B).
        
        Parameters
        ----------
        other : Sieve
            The sieve to subtract from this one.
        
        Returns
        -------
        set
            Elements in this sieve but not in the other.
        
        Examples
        --------
        >>> sieve1 = Sieve(2, 0, 10)  # Even numbers
        >>> sieve2 = Sieve(4, 0, 10)  # Multiples of 4
        >>> difference = sieve1 - sieve2
        >>> print(sorted(difference))
        [2, 6, 10]
        """
        return Operations.diff(self.S, other.S)
    
    def __xor__(self, other: 'Sieve') -> set:
        """
        Symmetric difference operation between two sieves (A ⊕ B).
        
        Parameters
        ----------
        other : Sieve
            The other sieve for symmetric difference.
        
        Returns
        -------
        set
            Elements in either sieve but not in both.
        
        Examples
        --------
        >>> sieve1 = Sieve(2, 0, 10)  # Even numbers
        >>> sieve2 = Sieve(3, 0, 10)  # Multiples of 3
        >>> sym_diff = sieve1 ^ sieve2
        >>> print(sorted(sym_diff))
        [2, 3, 4, 8, 9, 10]
        """
        return Operations.symm_diff(self.S, other.S)
    
    def __invert__(self) -> set:
        """
        Complement operation for the sieve (~A).
        
        Returns
        -------
        set
            All integers from 0 to N that are not in this sieve.
        
        Examples
        --------
        >>> sieve = Sieve(2, 0, 10)  # Even numbers
        >>> complement = ~sieve
        >>> print(sorted(complement))
        [1, 3, 5, 7, 9]
        """
        return self.compl
        

# ------------------------------------------------------------------------------
#  Generated Collection
# --------------

class GenCol:
    """
    Generated Collection - A multiplicative collection formed by repeatedly applying a generator.

    A multiplicative collection created by starting with an initial value and repeatedly
    multiplying by the generator, reducing modulo the period. This is the multiplicative
    analog to the Sieve class, useful for generating multiplicative pitch collections
    and harmonic series subsets.

    Parameters
    ----------
    generator : Union[str, int, float, Fraction]
        The generator value used for multiplication.
    period : Union[str, int, float, Fraction], optional
        The period of equivalence (default is 2).
    iterations : int, optional
        Number of times to apply the generator (default is 12).

    Attributes
    ----------
    generator : Fraction
        The generator value as a Fraction.
    period : Fraction
        The period of equivalence as a Fraction.
    iterations : int
        Number of generator applications.
    collection : list of Fraction
        The raw generated collection values.
    normalized_collection : list of Fraction
        Collection values normalized within the period.
    steps : set of Fraction
        Interval steps between consecutive normalized values.

    Examples
    --------
    Create a collection with default parameters (perfect fifth generator):
    
    >>> gen_col = GenCol()  # Uses '3/2' generator, period=2, iterations=7
    >>> print(*[str(f) for f in gen_col.normalized_collection])
    1 3/2 9/8 27/16 81/64 243/128 729/512 2187/2048
    """
    def __init__(self, generator: Union[str,int,float,Fraction] = '3/2', period: Union[str,int,float,Fraction] = 2, iterations: int = 7):
        """
        Initialize a Generated Collection.

        Parameters
        ----------
        generator : Union[str, int, float, Fraction], optional
            The generator value used for multiplication (default is '3/2').
        period : Union[str, int, float, Fraction], optional
            The period of equivalence (default is 2).
        iterations : int, optional
            Number of times to apply the generator (default is 7).
        """
        self._generator = Fraction(generator)
        self._period = Fraction(period)
        self._iterations = iterations
        self._collection = self._generate()
        
    @property
    def generator(self) -> Fraction:
        """
        The generator value as a Fraction.
        
        Returns
        -------
        Fraction
            The generator used for multiplication.
        """
        return self._generator
    
    @property
    def period(self) -> Fraction:
        """
        The period of equivalence as a Fraction.
        
        Returns
        -------
        Fraction
            The period within which values are considered equivalent.
        """
        return self._period
    
    @property
    def iterations(self) -> int:
        """
        Number of generator applications.
        
        Returns
        -------
        int
            The number of times the generator is applied.
        """
        return self._iterations
    
    @property
    def collection(self) -> List[Fraction]:
        """
        The raw generated collection values.
        
        Returns
        -------
        list of Fraction
            A copy of the generated collection values.
        """
        return self._collection.copy()
    
    @cached_property
    def normalized_collection(self) -> List[Fraction]:
        """
        Collection values normalized within the period.
        
        Returns
        -------
        list of Fraction
            The collection values reduced to within the period range.
        """
        normalized = []
        for value in self._collection:
            current = value
            while current >= self._period:
                current /= self._period
            normalized.append(current)
        return normalized
    
    @cached_property
    def steps(self) -> Set[Fraction]:
        """
        Interval steps between consecutive normalized values.
        
        Returns
        -------
        set of Fraction
            The intervals between consecutive values in the normalized collection.
        """
        values = sorted(self.normalized_collection)
        steps = set()
        
        for i in range(len(values)):
            if i < len(values) - 1:
                steps.add(values[i+1] / values[i])
            else:
                steps.add(self._period * values[0] / values[i])
        
        return steps
    
    def _generate(self) -> List[Fraction]:
        """
        Generate the collection by applying the generator iteratively.
        
        Returns
        -------
        list of Fraction
            The generated collection values.
        """
        return [self._generator ** i for i in range(self._iterations + 1)]
    
    def __str__(self):
        """
        String representation of the GenCol.
        
        Returns
        -------
        str
            Compact string representation showing generator, period, and iterations.
        """
        return f"GenCol(gen={self._generator}, period={self._period}, n={self._iterations})"
    
    def __repr__(self):
        """
        String representation of the GenCol.
        
        Returns
        -------
        str
            Compact string representation showing generator, period, and iterations.
        """
        return self.__str__()


# --------------------------------------------------------------------------------
# Combination Sets (CS)
# ---------------------

class CombinationSet:
    """
    A combinatorial set generating all r-combinations from a set of factors.

    This class generates all possible combinations of size r from a given set of factors,
    useful for combinatorial analysis and set operations. It also creates symbolic
    aliases and an associated graph structure for analysis.

    Parameters
    ----------
    factors : tuple, optional
        The factors to combine (default is ('A', 'B', 'C', 'D')).
    r : int, optional
        The size of each combination (default is 2).

    Attributes
    ----------
    factors : tuple
        The sorted tuple of input factors.
    rank : int
        The size of each combination.
    combos : set
        The set of all r-combinations from the factors.
    graph : networkx.Graph
        A complete graph with combinations as nodes.
    factor_to_alias : dict
        Mapping from factors to symbolic aliases.
    alias_to_factor : dict
        Mapping from symbolic aliases to factors.

    Examples
    --------
    Create combinations with default parameters:
    
    >>> cs = CombinationSet()  # Uses ('A', 'B', 'C', 'D') with r=2
    >>> print(cs.combos)
    {('A', 'B'), ('A', 'C'), ('A', 'D'), ('B', 'C'), ('B', 'D'), ('C', 'D')}
    """
    def __init__(self, factors:tuple = ('A', 'B', 'C', 'D'), r:int = 2):
        """
        Initialize a CombinationSet.

        Parameters
        ----------
        factors : tuple, optional
            The factors to combine (default is ('A', 'B', 'C', 'D')).
        r : int, optional
            The size of each combination (default is 2).
        """
        self._factors = tuple(sorted(factors))
        self._r = r
        self._combos = set(combinations(self._factors, self._r))
        self._factor_aliases = {f: sp.Symbol(chr(65 + i)) for i, f in enumerate(self._factors)}
        self._graph = self._generate_graph()

    @property
    def factors(self):
        """
        The sorted tuple of input factors.
        
        Returns
        -------
        tuple
            The factors used to generate combinations.
        """
        return self._factors
    
    @property
    def rank(self):
        """
        The size of each combination.
        
        Returns
        -------
        int
            The number of elements in each combination.
        """
        return self._r
    
    @property
    def combos(self):
        """
        The set of all r-combinations from the factors.
        
        Returns
        -------
        set
            All possible combinations of size r from the factors.
        """
        return self._combos
    
    @property
    def graph(self):
        """
        A complete graph with combinations as nodes.
        
        Returns
        -------
        networkx.Graph
            A complete graph where each node represents a combination.
        """
        return self._graph
    
    @property
    def factor_to_alias(self):
        """
        Mapping from factors to symbolic aliases.
        
        Returns
        -------
        dict
            Dictionary mapping each factor to a symbolic representation.
        """
        return self._factor_aliases
    
    @property
    def alias_to_factor(self):
        """
        Mapping from symbolic aliases to factors.
        
        Returns
        -------
        dict
            Dictionary mapping symbolic representations back to factors.
        """
        return {v: k for k, v in self._factor_aliases.items()}
    
    def _generate_graph(self):
        """
        Generate a complete graph with combinations as nodes.
        
        Returns
        -------
        networkx.Graph
            A complete graph where each node has a 'combo' attribute.
        """
        n_nodes = len(self._combos)
        G = nx.complete_graph(n_nodes)
        for i, combo in enumerate(self._combos):
            G.nodes[i]['combo'] = combo
        return G
    
    def __str__(self):    
        """
        String representation of the CombinationSet.
        
        Returns
        -------
        str
            A formatted string showing rank, factors, and combinations.
        """
        return (        
            f'Rank:    {self._r}\n'
            f'Factors: {self._factors}\n'
            f'Combos:  {self._combos}\n'
        )
    
    def __repr__(self) -> str:
        """
        String representation of the CombinationSet.
        
        Returns
        -------
        str
            A formatted string showing rank, factors, and combinations.
        """
        return self.__str__()


# ------------------------------------------------------------------------------
#  Partition Set
# --------------

class PartitionSet:
    """
    A set of integer partitions with associated graph structures for analysis.

    This class generates all partitions of an integer n into exactly k parts,
    analyzes their structural properties, and creates various graph representations
    for studying relationships between partitions.

    Parameters
    ----------
    n : int
        The integer to partition.
    k : int
        The number of parts in each partition.
    graph_type : {'feature_distance', 'decomposition_tree', 'substructure_embedding'}, optional
        The type of graph to construct (default is 'feature_distance').

    Attributes
    ----------
    data : pandas.DataFrame
        DataFrame containing partitions and their computed features.
    partitions : tuple
        Tuple of all generated partitions.
    mean : float
        The mean value of partition parts (n/k).
    graph : networkx.Graph or networkx.DiGraph
        Graph representation based on the specified graph_type.
    graph_type : str
        The type of graph construction used.

    Examples
    --------
    Generate partitions of 8 into 3 parts:
    
    >>> ps = PartitionSet(8, 3)
    >>> ps.data
       partition  unique_count  span  variance
    0  (6, 1, 1)             2     5    5.5556
    1  (5, 2, 1)             3     4    2.8889
    2  (4, 3, 1)             3     3    1.5556
    3  (4, 2, 2)             2     2    0.8889
    4  (3, 3, 2)             2     1    0.2222
    """
    def __init__(self, n: int, k: int, graph_type: Literal['feature_distance', 'decomposition_tree', 'substructure_embedding'] = 'feature_distance'):
        """
        Initialize a PartitionSet.

        Parameters
        ----------
        n : int
            The integer to partition.
        k : int
            The number of parts in each partition.
        graph_type : {'feature_distance', 'decomposition_tree', 'substructure_embedding'}, optional
            The type of graph to construct (default is 'feature_distance').
        """
        self._n = n
        self._k = k
        self._graph_type = graph_type
        self._data = self._generate_partitions()
        self._graph = self._generate_graph()
    
    def _generate_partitions(self) -> pd.DataFrame:
        """
        Generate all partitions of n into k parts with computed features.
        
        Returns
        -------
        pandas.DataFrame
            DataFrame with columns for partition, unique_count, span, and variance.
        """
        def backtrack(remaining: int, k: int, start: int, current: tuple) -> list:
            if k == 0:
                if remaining == 0:
                    return [{
                        'partition': current,
                        'unique_count': len(set(current)),
                        'span': max(current) - min(current),
                        'variance': np.var(current)
                    }]
                return []
            
            results = []
            for x in range(start, 0, -1):
                if x <= remaining:
                    results.extend(backtrack(remaining - x, k - 1, x, current + (x,)))
            return results
                    
        return pd.DataFrame(backtrack(self._n, self._k, self._n, ()))
    
    def _generate_graph(self) -> nx.Graph:
        """
        Generate a graph based on the specified graph type.
        
        Returns
        -------
        networkx.Graph or networkx.DiGraph
            The generated graph representation of partition relationships.
        
        Raises
        ------
        ValueError
            If an unknown graph type is specified.
        """
        if self._graph_type == 'feature_distance':
            return self._build_feature_distance_graph()
        elif self._graph_type == 'decomposition_tree':
            return self._build_decomposition_tree_graph()
        elif self._graph_type == 'substructure_embedding':
            return self._build_substructure_embedding_graph()
        else:
            raise ValueError(f"Unknown graph type: {self._graph_type}")
    
    def _build_feature_distance_graph(self) -> nx.Graph:
        """
        Build a graph where edge weights represent feature distances.
        
        Returns
        -------
        networkx.Graph
            A complete graph with edge weights based on normalized feature distances.
        """
        features = self._data[['unique_count', 'span', 'variance']].values
        features_normalized = (features - features.mean(axis=0)) / (features.std(axis=0) + 1e-8)
        distances = pdist(features_normalized, metric='euclidean')
        distance_matrix = squareform(distances)
        
        G = nx.Graph()
        n_partitions = len(self._data)
        
        for i in range(n_partitions):
            G.add_node(i, partition=tuple(self._data.iloc[i]['partition']))
        
        for i in range(n_partitions):
            for j in range(i + 1, n_partitions):
                G.add_edge(i, j, weight=distance_matrix[i, j])
        
        return G
    
    def _build_decomposition_tree_graph(self) -> nx.Graph:
        """
        Build a graph where edge weights represent transformation costs.
        
        Returns
        -------
        networkx.Graph
            A complete graph with edge weights based on partition transformation costs.
        """
        partitions = [tuple(row['partition']) for _, row in self._data.iterrows()]
        G = nx.Graph()
        
        for i, partition in enumerate(partitions):
            G.add_node(i, partition=partition)
        
        for i, partition_a in enumerate(partitions):
            for j, partition_b in enumerate(partitions):
                if i < j:
                    cost = self._partition_transformation_cost(partition_a, partition_b)
                    G.add_edge(i, j, weight=cost)
        
        return G
    
    def _partition_transformation_cost(self, partition_a: tuple, partition_b: tuple) -> float:
        """
        Calculate the cost of transforming one partition to another.
        
        Parameters
        ----------
        partition_a : tuple
            The first partition.
        partition_b : tuple
            The second partition.
        
        Returns
        -------
        float
            The transformation cost between the partitions.
        """
        a_parts = sorted(partition_a, reverse=True)
        b_parts = sorted(partition_b, reverse=True)
        
        max_len = max(len(a_parts), len(b_parts))
        a_padded = a_parts + [0] * (max_len - len(a_parts))
        b_padded = b_parts + [0] * (max_len - len(b_parts))
        
        cost = sum(abs(a - b) for a, b in zip(a_padded, b_padded)) / 2.0
        
        return cost
    
    def _build_substructure_embedding_graph(self) -> nx.DiGraph:
        """
        Build a directed graph based on substructure embedding costs.
        
        Returns
        -------
        networkx.DiGraph
            A directed graph where edge weights represent embedding costs.
        """
        partitions = [tuple(row['partition']) for _, row in self._data.iterrows()]
        G = nx.DiGraph()
        
        for i, partition in enumerate(partitions):
            G.add_node(i, partition=partition)
        
        for i, partition_a in enumerate(partitions):
            for j, partition_b in enumerate(partitions):
                if i != j:
                    embedding_cost = self._substructure_embedding_cost(partition_a, partition_b)
                    G.add_edge(i, j, weight=embedding_cost)
        
        return G
    
    def _substructure_embedding_cost(self, partition_a: tuple, partition_b: tuple) -> float:
        """
        Calculate the cost of embedding one partition structure into another.
        
        Parameters
        ----------
        partition_a : tuple
            The partition to embed.
        partition_b : tuple
            The target partition for embedding.
        
        Returns
        -------
        float
            The embedding cost between the partitions.
        """
        a_parts = sorted(partition_a, reverse=True)
        b_parts = sorted(partition_b, reverse=True)
        
        b_remaining = list(b_parts)
        total_cost = 0.0
        
        for a_part in a_parts:
            best_match_cost = float('inf')
            best_match_idx = -1
            
            for idx, b_part in enumerate(b_remaining):
                if b_part >= a_part:
                    cost = abs(b_part - a_part) / max(a_part, 1)
                    if cost < best_match_cost:
                        best_match_cost = cost
                        best_match_idx = idx
                else:
                    cost = 2.0 + (a_part - b_part) / a_part
                    if cost < best_match_cost:
                        best_match_cost = cost
                        best_match_idx = idx
            
            if best_match_idx != -1:
                total_cost += best_match_cost
                if b_remaining[best_match_idx] >= a_part:
                    b_remaining[best_match_idx] -= a_part
                    if b_remaining[best_match_idx] == 0:
                        b_remaining.pop(best_match_idx)
                else:
                    b_remaining.pop(best_match_idx)
            else:
                total_cost += 5.0
        
        return total_cost
    
    @property
    def data(self):
        """
        DataFrame containing partitions and their computed features.
        
        Returns
        -------
        pandas.DataFrame
            DataFrame with partition, unique_count, span, and variance columns.
        """
        return self._data
    
    @property
    def partitions(self):
        """
        Tuple of all generated partitions.
        
        Returns
        -------
        tuple
            All partitions of n into k parts.
        """
        return tuple(self._data['partition'])
    
    @property
    def mean(self) -> float:
        """
        The mean value of partition parts.
        
        Returns
        -------
        float
            The arithmetic mean n/k.
        """
        return self._n / self._k
    
    @property
    def graph(self):
        """
        Graph representation based on the specified graph_type.
        
        Returns
        -------
        networkx.Graph or networkx.DiGraph
            The graph representation of partition relationships.
        """
        return self._graph
    
    @property
    def graph_type(self):
        """
        The type of graph construction used.
        
        Returns
        -------
        str
            The graph type identifier.
        """
        return self._graph_type

    def __str__(self) -> str:
        """
        String representation of the PartitionSet.
        
        Returns
        -------
        str
            A formatted string showing partition data and statistics.
        """
        display_df = self._data.copy()
        display_df['variance'] = display_df['variance'].round(4)
        
        df_str = str(display_df)
        width = max(len(line) for line in df_str.split('\n'))
        border = '-' * width
        
        header = (
            f"{border}\n"
            f"PS(n={self._n}, k={self._k}) - Graph: {self._graph_type}\n"
            f"Mean: ~{round(self.mean, 4)}\n"
            f"{border}\n"
        )
        return header + df_str + f"\n{border}\n"
    
    def __repr__(self) -> str:
        """
        String representation of the PartitionSet.
        
        Returns
        -------
        str
            A formatted string showing partition data and statistics.
        """
        return self.__str__()




from fractions import Fraction
from typing import TypeVar, cast, Optional, Union, List
from ..pitch import EquaveCyclicCollection, AddressedPitchCollection, IntervalType, _addressed_collection_cache, IntervalList, Pitch
import numpy as np
from ..utils.interval_normalization import equave_reduce

PC = TypeVar('PC', bound='Scale')

class AddressedScale(AddressedPitchCollection):
    """
    A scale bound to a specific root pitch.
    
    AddressedScale provides access to the actual pitches of a scale when rooted
    at a specific pitch, enabling work with concrete frequencies and pitch names
    rather than abstract intervals.
    
    Examples:
        >>> from klotho.tonos import Scale
        >>> major = Scale(["1/1", "9/8", "5/4", "4/3", "3/2", "5/3", "15/8"])
        >>> c_major = major.root("C4")
        >>> c_major[0]
        C4
        >>> c_major[2] 
        E4
    """
    pass

class Scale(EquaveCyclicCollection[IntervalType]):
    """
    A musical scale with automatic sorting, deduplication, and equave removal.
    
    Scale represents a collection of pitch intervals that form a musical scale.
    It automatically sorts degrees, removes duplicates, removes the equave interval,
    and ensures the unison (1/1) is present. Scales support infinite equave 
    displacement for accessing pitches in different octaves.
    
    Args:
        degrees: List of intervals as ratios, decimals, or numbers
        equave: The interval of equivalence, defaults to "2/1" (octave)
        
    Examples:
        >>> scale = Scale(["1/1", "9/8", "5/4", "4/3", "3/2", "5/3", "15/8"])
        >>> scale.degrees
        [Fraction(1, 1), Fraction(9, 8), Fraction(5, 4), Fraction(4, 3), Fraction(3, 2), Fraction(5, 3), Fraction(15, 8)]
        
        >>> scale[7]  # Next octave
        Fraction(2, 1)
        
        >>> scale.mode(1)  # Dorian mode
        Scale([Fraction(1, 1), Fraction(10, 9), ...], equave=2)
        
        >>> c_major = scale.root("C4")
        >>> c_major[0]
        C4
    """
    def __init__(self, degrees: IntervalList = ["1/1", "9/8", "5/4", "4/3", "3/2", "5/3", "15/8"], 
                 equave: Optional[Union[float, Fraction, int, str]] = "2/1"):
        super().__init__(degrees, equave)
        self._mode_cache = {}
        
    def _process_degrees(self, degrees: IntervalList) -> List[IntervalType]:
        processed = super()._process_degrees(degrees)
        
        if not processed:
            return processed
        
        if self._interval_type == float:
            if not processed or abs(processed[0]) >= 1e-6:
                processed.insert(0, 0.0)
        else:
            if not processed or processed[0] != Fraction(1, 1):
                processed.insert(0, Fraction(1, 1))
        
        return processed
        
    def mode(self, mode_number: int) -> 'Scale':
        if mode_number in self._mode_cache:
            return self._mode_cache[mode_number]
            
        if mode_number == 0:
            return self
            
        size = len(self._degrees)
        if size == 0:
            return Scale([], self._equave)
        
        start_index = mode_number % size
        if start_index < 0:
            start_index += size
        
        first_degree = self._degrees[start_index]
        modal_degrees = []
        
        if self.interval_type == float:
            for i in range(size):
                current_idx = (start_index + i) % size
                
                if i == 0:
                    modal_degrees.append(0.0)
                else:
                    interval = self._degrees[current_idx] - first_degree
                    if current_idx < start_index:
                        equave_value = self._equave if isinstance(self._equave, float) else 1200 * np.log2(float(self._equave))
                        interval += equave_value
                    modal_degrees.append(interval)
        else:
            for i in range(size):
                current_idx = (start_index + i) % size
                
                if i == 0:
                    modal_degrees.append(Fraction(1, 1))
                else:
                    interval = self._degrees[current_idx] / first_degree
                    if current_idx < start_index:
                        equave_value = self._equave if isinstance(self._equave, Fraction) else Fraction.from_float(2 ** (self._equave / 1200))
                        interval *= equave_value
                    modal_degrees.append(interval)
        
        result = Scale(modal_degrees, self._equave)
        self._mode_cache[mode_number] = result
        return result
        
    def __invert__(self: PC) -> PC:
        if self.interval_type == float:
            inverted = [0.0 if abs(interval) < 1e-6 else 1200.0 - interval for interval in self._degrees]
        else:
            inverted = [Fraction(1, 1) if interval == Fraction(1, 1) else Fraction(interval.denominator * 2, interval.numerator) for interval in self._degrees]
        
        return Scale(sorted(inverted), self._equave)
    
    def __neg__(self: PC) -> PC:
        return self.__invert__()
        
    def root(self, other: Union[Pitch, str]) -> 'AddressedScale':
        if isinstance(other, str):
            other = Pitch(other)
            
        cache_key = (id(self), id(other))
        if cache_key not in _addressed_collection_cache:
            _addressed_collection_cache[cache_key] = AddressedScale(self, other)
        return _addressed_collection_cache[cache_key] 
    


from typing import Tuple, List, Union
import networkx as nx
import pandas as pd
from ..graphs import Graph


class Lattice(Graph):
    """
    A generic n-dimensional lattice structure.
    
    A lattice provides a discrete sampling of n-dimensional space with integer
    coordinates. This can be used for various applications such as spatial fields,
    microtonal tone lattices, or any discrete grid-based structure.
    
    Parameters
    ----------
    dimensionality : int
        Number of dimensions.
    resolution : int or list of int
        Number of points along each dimension, or list of resolutions per dimension.
    bipolar : bool, optional
        If True, coordinates range from -resolution to +resolution. 
        If False, coordinates range from 0 to resolution (default is True).
    """
    
    def __init__(self, 
                 dimensionality : int                   = 2, 
                 resolution     : Union[int, List[int]] = 10, 
                 bipolar        : bool                  = True,
                 periodic       : bool                  = False,
        ):
        
        self._dimensionality = dimensionality
        self._bipolar = bipolar
        
        if isinstance(resolution, int):
            self._resolution = [resolution] * dimensionality
        else:
            if len(resolution) != dimensionality:
                raise ValueError(f"Resolution list length {len(resolution)} must match dimensionality {dimensionality}")
            self._resolution = resolution
        
        if self._bipolar:
            dims = [range(-res, res + 1) for res in self._resolution]
        else:
            dims = [range(0, res + 1) for res in self._resolution]
        
        lattice_graph = nx.grid_graph(dims, periodic=periodic)
        
        # Initialize Graph components manually since coordinate tuples aren't integers
        self._graph = lattice_graph
        self._meta = pd.DataFrame(index=[''])
        self._next_id = 0  # Not used since we use coordinate tuples as node IDs
    
    @property
    def dimensionality(self) -> int:
        """
        Number of dimensions in the lattice.
        
        Returns
        -------
        int
            The dimensionality of the lattice.
        """
        return self._dimensionality
    
    @property
    def resolution(self) -> List[int]:
        """
        Resolution along each dimension.
        
        Returns
        -------
        list of int
            Copy of the resolution list to prevent external modification.
        """
        return self._resolution.copy()
    
    @property
    def bipolar(self) -> bool:
        """
        Whether the lattice uses bipolar coordinates.
        
        Returns
        -------
        bool
            True if coordinates range from -resolution to +resolution,
            False if coordinates range from 0 to resolution.
        """
        return self._bipolar  
    
    def get_neighbors(self, coordinate: Tuple[int, ...]) -> List[Tuple[int, ...]]:
        """
        Get the neighboring coordinates of a given coordinate.
        
        Parameters
        ----------
        coordinate : tuple of int
            The lattice coordinate.
            
        Returns
        -------
        list of tuple of int
            List of neighboring coordinates.
            
        Raises
        ------
        KeyError
            If coordinate is not in lattice.
        """
        if coordinate not in self.nodes:
            raise KeyError(f"Coordinate {coordinate} not in lattice")
        return list(self.graph.neighbors(coordinate))
    
    def coords(self) -> List[Tuple[int, ...]]:
        """
        Get all coordinates in the lattice.
        
        Returns
        -------
        list of tuple of int
            List of all lattice coordinates.
        """
        return list(self.nodes)    
    
    def __str__(self) -> str:
        """String representation of the lattice."""
        return (f"Lattice(dimensionality={self._dimensionality}, "
                f"resolution={self._resolution}, "
                f"bipolar={self._bipolar}, "
                f"coordinates={len(self.nodes)})")
    
    def __repr__(self) -> str:
        return self.__str__() 


import numpy as np
from collections.abc import Iterable
from itertools import cycle
from math import lcm

__all__ = [
    'Norg',
    'Pattern',
]

class Norg:
    '''
    Class for Per Nørgård's "infinity" sequences.

    see: https://web.archive.org/web/20071010091253/http://www.pernoergaard.dk/eng/strukturer/uendelig/uindhold.html
    '''
    @staticmethod
    def inf(start: int = 0, size: int = 128, step:int = 1):
        '''
        from: https://web.archive.org/web/20071010092334/http://www.pernoergaard.dk/eng/strukturer/uendelig/ukonstruktion05.html

        '''
        if start == 0 and step == 1:
            p = np.empty(size, dtype=int)
            p[0] = 0
            p[1] = 1
            for i in range(1, (size - 1) // 2 + 1):
                delta = p[i] - p[i - 1]
                if 2 * i < size:
                    p[2 * i] = p[2 * i - 2] - delta
                if 2 * i + 1 < size:
                    p[2 * i + 1] = p[2 * i - 1] + delta
            return p
        return np.array([Norg.inf_num(start + step * i) for i in range(size)])

    @staticmethod
    def inf_num(n):
        '''
        see: https://arxiv.org/pdf/1402.3091.pdf

        '''
        if n == 0: return 0
        if n % 2 == 0:
            return -Norg.inf_num(n // 2)
        return Norg.inf_num((n - 1) // 2) + 1

    @staticmethod
    def n_partite(seed: list = [0,-2,-1], inv_pat: list = [-1,1,1], size: int = 128):
        '''
        Generalized form of the tripartite series for any arbitrary length seed and inv_pat.

        from: https://web.archive.org/web/20071010091606/http://www.pernoergaard.dk/eng/strukturer/uendelig/u3.html

        '''
        seed_len = len(seed)
        p = np.empty(size, dtype=int)
        p[:seed_len] = seed
        inv_pat_cycle = cycle(inv_pat)
        for i in range(1, (size - 1) // seed_len + 1):
            delta = p[i] - p[i - 1]
            for j in range(seed_len):
                if seed_len * i + j < size:
                    p[seed_len * i + j] = p[seed_len * i + j - seed_len] + next(inv_pat_cycle) * delta
        return p
    
    @staticmethod
    def lake():
        '''
        see: https://web.archive.org/web/20071010093955/http://www.pernoergaard.dk/eng/strukturer/toneso/tkonstruktion.html

        '''
        pass


class Pattern:
    def __init__(self, iterable, end=False):
        self._iterable = iterable
        self._cycles, self._pattern_length = self._create_cycles(iterable)
        self._end = end
        self._current = 0
        
    def _create_cycles(self, item):
        if isinstance(item, Iterable) and not isinstance(item, (str, bytes)):
            sub_results = [self._create_cycles(subitem) for subitem in item]
            subcycles, lengths = zip(*sub_results)
            this_level_length = len(item)
            nested_lengths = [this_level_length * length for length in lengths if length > 0]
            total_length = lcm(*nested_lengths) if nested_lengths else this_level_length
            return cycle(subcycles), total_length
        return item, 1

    def __iter__(self):
        return self

    def __next__(self):
        if self._current >= self._pattern_length and self._end:
            return next(self._end) if isinstance(self._end, Pattern) else self._end
        self._current += 1
        return self._get_next(self._cycles)

    def _get_next(self, cyc):
        item = next(cyc)
        while isinstance(item, cycle):
            item = self._get_next(item)
        return item
    
    def __len__(self):
        return self._pattern_length
    
    def __str__(self):
        return str(list(self._iterable))

    def __repr__(self):
        return self.__str__()



"""
Scheduler: A module for scheduling musical events and nodes.

Provides the Scheduler class for managing timed musical events 
with priority-based scheduling.
"""

from uuid import uuid4
import heapq
import json
import os
from typing import Union


class Scheduler:
    def __init__(self):
        self.events = []
        self.total_events = 0
        self.event_counter = 0  # sorting for final tiebreaker
        
    def new_node(self, synth_name: str, start: float = 0, dur: Union[float, None] = None, group: str = None, **pfields):
        uid = str(uuid4()).replace('-', '')
        
        event = {
            "type": "new",
            "id": uid,
            "synthName": synth_name,
            "start": start,
            "pfields": pfields
        }
        
        if group:
            event["group"] = group
        else:
            event["group"] = "default"
            
        priority = 0 # higher priority
        heapq.heappush(self.events, (start, priority, uid, self.event_counter, event))
        self.event_counter += 1
        self.total_events += 1
        
        if dur:
            self.set_node(uid, start = start + dur, gate = 0)
        
        return uid

    def set_node(self, uid: str, start: float, **pfields):
        event = {
            "type": "set",
            "id": uid,
            "start": start,
            "pfields": pfields
        }
        
        priority = 1 # lower priority
        heapq.heappush(self.events, (start, priority, uid, self.event_counter, event))
        self.event_counter += 1
        self.total_events += 1
    
    def free_node(self, uid: str):
        event = {
            "type": "free",
            "id": uid
        }
        heapq.heappush(self.events, (0, 0, uid, self.event_counter, event))
        self.event_counter += 1
        self.total_events += 1
        
    def clear_events(self):
        self.events = []
        self.total_events = 0
        self.event_counter = 0
        
    def write(self, filepath, start_time: Union[float, None] = None, time_scale: float = 1.0):
        sorted_events = []
        events_copy = self.events.copy()
        
        if events_copy:
            if start_time is not None:
                min_start = min(start for start, _, _, _, _ in events_copy)
                time_shift = start_time - min_start
            else:
                time_shift = 0
            
            while events_copy:
                start, _, _, _, event = heapq.heappop(events_copy)
                new_start = (start + time_shift) * time_scale
                event["start"] = new_start
                sorted_events.append(event)
            
        try:
            with open(filepath, 'w') as f:
                json.dump(sorted_events, f, indent=2)
            print(f"Successfully wrote {self.total_events} events to {os.path.abspath(filepath)}")
        except Exception as e:
            print(f"Error writing to {filepath}: {e}") 


from klotho.topos.graphs import Tree
from .algorithms import *
from klotho.tonos.utils.interval_normalization import reduce_interval
from typing import Tuple, Union
from fractions import Fraction

class HarmonicTree(Tree):
    '''
    '''
    def __init__(self,
                 root:int                                  = 1,
                 children:Tuple[int, ...]                  = (1,),
                 equave:Union[Fraction,int,float,str,None] = None,
                 span:int                                  = 1):
        
        super().__init__(root, children)
        
        self._meta['equave'] = Fraction(equave) if equave is not None else None
        self._meta['span']   = span
        
        self._evaluate()

    def _evaluate(self):
        self.graph.nodes[0]['multiple'] = self._graph.nodes[self.root]['label']
        self.graph.nodes[0]['harmonic'] = self._graph.nodes[self.root]['label']
        def process_subtree(node=0, factor=1):
            value = self.graph.nodes[node]['label']
            self.graph.nodes[node]['multiple'] = value if value > 0 else Fraction(1, abs(value))
            children = list(self.graph.successors(node))
            
            if not children:
                harmonic = value * factor
                self.graph.nodes[node]['harmonic'] = harmonic
                
                if self.equave is not None:
                    self.graph.nodes[node]['ratio'] = reduce_interval(Fraction(harmonic), self.equave, self.span)
                else:
                    self.graph.nodes[node]['ratio'] = harmonic
                
                return
            else:
                for child in children:
                    value = self.graph.nodes[child]['label']
                    self.graph.nodes[child]['multiple'] = value
                    harmonic = value * factor
                    self.graph.nodes[child]['harmonic'] = harmonic
                    
                    if self.equave is not None:
                        self.graph.nodes[child]['ratio'] = reduce_interval(Fraction(harmonic), self.equave, self.span)
                    else:
                        self.graph.nodes[child]['ratio'] = harmonic
                    
                    if self.graph.out_degree(child) > 0:
                        process_subtree(child, harmonic)
        
        process_subtree()

    @property
    def harmonics(self):
        return tuple(self.graph.nodes[n]['harmonic'] for n in self.leaf_nodes)
    
    @property
    def ratios(self):
        return tuple(self.graph.nodes[n]['ratio'] for n in self.leaf_nodes)
    
    @property
    def equave(self):
        return self._meta['equave'].iloc[0]
    
    @property
    def span(self):
        return self._meta['span'].iloc[0]
    
    # @property
    # def inverse(self):
    #     return HarmonicTree(
    #         root     = self.__root,
    #         children = self.__children,
    #         equave   = self.__equave,
    #         span = self.__span,
    #         inverse  = not self.__inverse
    #     )
    


from ..trees import Tree
import networkx as nx
from typing import Tuple
from itertools import count

__all__ = [
    'factor_children',
    'refactor_children',
    'get_signs',
    'get_abs',
    'rotate_children',
    'rotate_tree',
    'print_subdivisons',
    'prune_to_depth',
    'prune_leaves',
    'path_to_node',
    'extract_subtree',
    'are_isomorphic',
    'get_levels'
]

def factor_children(subdivs:tuple) -> tuple:
    def _factor(subdivs, acc):
        for element in subdivs:
            if isinstance(element, tuple):
                _factor(element, acc)
            else:
                acc.append(element)
        return acc
    return tuple(_factor(subdivs, []))

def refactor_children(subdivs:tuple, factors:tuple) -> tuple:
    def _refactor(subdivs, index):
        result = []
        for element in subdivs:
            if isinstance(element, tuple):
                nested_result, index = _refactor(element, index)
                result.append(nested_result)
            else:
                result.append(factors[index])
                index += 1
        return tuple(result), index
    return _refactor(subdivs, 0)[0]

def get_signs(subdivs):
        signs = []
        for element in subdivs:
            if isinstance(element, tuple):
                signs.extend(get_signs(element))
            else:
                signs.append(1 if element >= 0 else -1)
        return signs
    
def get_abs(subdivs):
        result = []
        for element in subdivs:
            if isinstance(element, tuple):
                result.extend(get_abs(element))
            else:
                result.append(abs(element))
        return result

def rotate_children(subdivs: tuple, n: int = 1, preserve_signs: bool = False) -> tuple:
    """Rotates the children of a nested tuple structure.
    
    Args:
        subdivs: Nested tuple structure to rotate
        n: Number of positions to rotate
        preserve_signs: If True, preserves the signs of numbers while rotating their absolute values
    """
    if not preserve_signs:
        factors = factor_children(subdivs)
        n = n % len(factors)
        factors = factors[n:] + factors[:n]
        return refactor_children(subdivs, factors)
    
    signs = get_signs(subdivs)
    abs_values = get_abs(subdivs)
    
    n = n % len(abs_values)
    rotated_values = abs_values[n:] + abs_values[:n]
    
    signed_values = [val * sign for val, sign in zip(rotated_values, signs)]
    
    return refactor_children(subdivs, tuple(signed_values))

def rotate_tree(tree:Tree, n:int=1) -> Tree:
    return Tree(tree.group.D, rotate_children(tree.group.S, n))

def print_subdivisons(subdivs):
    """Format nested tuple structure removing commas."""
    if isinstance(subdivs, (tuple, list)):
        inner = ' '.join(str(print_subdivisons(x)) for x in subdivs)
        return f"({inner})"
    return str(subdivs)

def prune_to_depth(tree: Tree, max_depth: int) -> Tree:
    """Prunes the tree to a maximum depth, removing all nodes beyond that depth."""
    if max_depth < 0:
        raise ValueError("max_depth must be non-negative")
    
    G = tree.graph.copy()
    depths = nx.single_source_shortest_path_length(G, 0)
    nodes_to_remove = [n for n, depth in depths.items() if depth > max_depth]
    G.remove_nodes_from(nodes_to_remove)
    
    return Tree.from_graph(G)

def prune_leaves(tree: Tree, n: int) -> Tree:
    """Prunes n levels from each branch, starting from the leaves."""
    if n < 0:
        raise ValueError("n must be non-negative")
    if n == 0:
        return tree
    
    G = tree.graph.copy()
    for _ in range(n):
        leaves = [node for node in G.nodes() if G.out_degree(node) == 0]
        G.remove_nodes_from(leaves)
        if len(G) == 1:  # Only root remains
            break
    
    return Tree.from_graph(G)

def path_to_node(tree: Tree, node_id: int) -> list[int]:
    """Returns the path from root to the specified node as a list of node IDs."""
    if node_id not in tree.graph:
        raise ValueError(f"Node {node_id} not found in tree")
    
    try:
        path = nx.shortest_path(tree.graph, 0, node_id)
        return path
    except nx.NetworkXNoPath:
        raise ValueError(f"No path exists to node {node_id}")

def extract_subtree(tree: Tree, root_id: int) -> Tree:
    """Creates a new tree from the subtree rooted at the specified node."""
    if root_id not in tree.graph:
        raise ValueError(f"Node {root_id} not found in tree")
    
    G = tree.graph.copy()
    descendants = nx.descendants(G, root_id)
    descendants.add(root_id)
    subtree = G.subgraph(descendants).copy()
    
    # Relabel nodes to ensure root is 0
    mapping = {root_id: 0}
    counter = count(1)
    for node in subtree.nodes():
        if node != root_id:
            mapping[node] = next(counter)
    subtree = nx.relabel_nodes(subtree, mapping)
    
    return Tree.from_graph(subtree)

def are_isomorphic(tree1: Tree, tree2: Tree) -> bool:
    """Returns True if the two trees are structurally identical."""
    return nx.is_isomorphic(tree1.graph, tree2.graph, 
                           node_match=lambda x, y: x['label'] == y['label'])

def get_levels(tree: Tree) -> list[list[int]]:
    """Returns lists of nodes grouped by their depth in the tree."""
    depths = nx.single_source_shortest_path_length(tree.graph, 0)
    max_depth = max(depths.values())
    
    levels = [[] for _ in range(max_depth + 1)]
    for node, depth in depths.items():
        levels[depth].append(node)
    
    return levels




"""
Streamer: A module for real-time OSC streaming of musical events.

Provides StreamBuilder and Streamer classes for real-time communication
with external audio systems via OSC protocol.
"""

from pythonosc import udp_client, osc_message_builder, dispatcher, osc_server
from uuid import uuid4
import threading
import time
from tqdm import tqdm
import heapq


class StreamBuilder:
    def __init__(self, scheduler: 'Streamer', event_type: str, uid: str, synth_name: str, start: float, params: dict):
        self.scheduler = scheduler
        self.event_type = event_type
        self.uid = str(uuid4()).replace('-', '') if uid is None else uid
        self.synth_name = synth_name
        self.start = start
        self.params = params
        self._add_event()
            
    def _add_event(self):
        args = [self.uid, self.synth_name, self.start] + [item for sublist in self.params.items() for item in sublist]
        event_type = self.event_type
        heapq.heappush(self.scheduler.events, (self.start, (event_type, args)))
        self.scheduler.total_events += 1
    
    def __str__(self):
        return self.uid
        
    def __repr__(self):
        return self.uid


class Streamer:
    def __init__(self, ip:str='127.0.0.1', send_port:int=57121, receive_port:int=9000):
        self.client = udp_client.SimpleUDPClient(ip, send_port)
        self.events = []
        self.paused = True
        self.events_processed = 0
        self.total_events = 0
        self.events_sent = 0
        self.batch_size = 25

        self.dispatcher = dispatcher.Dispatcher()
        self.dispatcher.map("/pause", self.pause_handler)
        self.dispatcher.map("/resume", self.resume_handler)
        self.dispatcher.map("/event_processed", self.event_processed_handler)
        self.dispatcher.map("/reset", self.reset_handler)
        self.dispatcher.map("/start", self.start)
        self.dispatcher.map("/new_synth", self.new_synth_handler)
        self.dispatcher.map("/set_synth", self.set_synth_handler)
        self.dispatcher.map("/clear_events", self.clear_events)
        
        self.server = osc_server.ThreadingOSCUDPServer((ip, receive_port), self.dispatcher)
        self.server_thread = None
        self.running = False
        self.send_progress = None
        self.process_progress = None
        
    def init_progress_bars(self):
        if self.send_progress:
            self.send_progress.close()
        if self.process_progress:
            self.process_progress.close()
            
        self.send_progress = tqdm(total=self.total_events, 
                                desc="Ready to send", 
                                unit="events",
                                position=0,
                                leave=True)
        
        self.process_progress = tqdm(total=self.total_events,
                                   desc="Ready to process",
                                   unit="events",
                                   position=1,
                                   leave=True)
    
    def reset_progress_bars(self):
        if self.send_progress:
            self.send_progress.clear()
            self.send_progress.reset(total=self.total_events)
            self.send_progress.set_description("Ready to send")
        if self.process_progress:
            self.process_progress.clear()
            self.process_progress.reset(total=self.total_events)
            self.process_progress.set_description("Ready to process")
        
    def pause_handler(self, address, *args):
        if not self.paused:
            self.paused = True
            if self.send_progress:
                self.send_progress.set_description("Paused sending")

    def resume_handler(self, address, *args):
        self.paused = False
        if self.send_progress:
            self.send_progress.set_description("Sending")

    def event_processed_handler(self, address, *args):
        self.events_processed += 1
        if self.process_progress:
            self.process_progress.update(1)
        if self.events_sent == self.total_events:
            if self.send_progress:
                self.send_progress.set_description("All Events Sent")
        if self.events_processed == self.total_events:
            if self.process_progress:
                self.process_progress.set_description("All Events Processed")

    def reset_handler(self, address, *args):
        self.events_processed = 0
        self.events_sent = 0
        self.paused = True
        self.reset_progress_bars()

    def new_synth(self, synth_name:str = None, start:float = 0, **params):
        return StreamBuilder(self, 'new', None, synth_name, start, params).uid

    def set_synth(self, uid:str, start:float, **params):
        StreamBuilder(self, 'set', uid, None, start, params)

    def new_synth_handler(self, address, uid, synth_name, start, *params):
        param_dict = {params[i]: params[i+1] for i in range(0, len(params), 2)}
        StreamBuilder(self, 'new', uid, synth_name, float(start), param_dict)
        self.reset_progress_bars()
        
    def set_synth_handler(self, address, uid, start, *params):
        param_dict = {params[i]: params[i+1] for i in range(0, len(params), 2)}
        StreamBuilder(self, 'set', uid, None, float(start), param_dict)
        self.reset_progress_bars()
        
    def start(self, address, *args):
        if self.events_sent == self.total_events:
            self.events_processed = 0
            self.events_sent = 0
            self.reset_progress_bars()
        self.paused = False
        self.send_progress.set_description("Sending")
        self.process_progress.set_description("Processing")
        self.send_events()

    def send_events(self):
        events = self.events.copy()
        while events:
            while self.paused:
                time.sleep(0.01)
            _, (event_type, content) = heapq.heappop(events)
            
            msg = osc_message_builder.OscMessageBuilder(address='/storeEvent')
            msg.add_arg(event_type)
            for item in content:
                msg.add_arg(item)
            self.client.send(msg.build())
            self.events_sent += 1
            self.send_progress.update(1)
            self.process_progress.total = self.events_sent
            self.process_progress.refresh()
            time.sleep(0.01)

        eot_msg = osc_message_builder.OscMessageBuilder(address='/storeEvent')
        eot_msg.add_arg('end_of_transmission')
        self.client.send(eot_msg.build())
        self.send_progress.set_description("All Events Sent")

    def clear_events(self, address=None, *args):
        self.events = []
        self.total_events = 0
        self.events_processed = 0
        self.events_sent = 0
        self.paused = True
        if self.send_progress:
            self.send_progress.reset()
        if self.process_progress:
            self.process_progress.reset()

    def stop_server(self):
        if self.running:
            self.server.shutdown()
            if self.server_thread and self.server_thread.is_alive():
                self.server_thread.join()
            self.server_thread = None
            self.running = False
            if self.send_progress:
                self.send_progress.close()
            if self.process_progress:
                self.process_progress.close()

    def run(self):
        if not self.running:
            self.running = True
            self.server_thread = threading.Thread(target=self.server.serve_forever)
            self.server_thread.start()
            self.init_progress_bars()
        else:
            print("Server is already running.") 


from typing import List, Callable, TypeVar, Any, Tuple
import numpy as np
import networkx as nx

T = TypeVar('T')

def cost_matrix(items: List[T], cost_function: Callable[[T, T], float], **kwargs: Any) -> Tuple[np.ndarray, List[T]]:
    """
    Generate a symmetric cost matrix for a collection of items.
    
    Create a numpy array representing pairwise costs between items using
    a provided cost function. The resulting matrix is symmetric with
    indices corresponding to item positions in the input list.

    Parameters
    ----------
    items : List[T]
        List of items to compute pairwise costs for. Items can be of any
        type that the cost function can handle.
    cost_function : Callable[[T, T], float]
        Function that takes two items and returns a numeric cost value.
        Should be symmetric (cost(a, b) == cost(b, a)) for best results.
    **kwargs : Any
        Additional keyword arguments to pass to the cost function.

    Returns
    -------
    Tuple[numpy.ndarray, List[T]]
        A tuple containing:
        - Symmetric cost matrix as numpy array where entry (i, j) 
          contains cost_function(items[i], items[j])
        - List of items in the same order as matrix indices

    Examples
    --------
    Create a distance matrix for 2D points:
    
    >>> import math
    >>> points = [(0, 0), (1, 1), (2, 0)]
    >>> def euclidean_distance(p1, p2):
    ...     return math.sqrt((p1[0] - p2[0])**2 + (p1[1] - p2[1])**2)
    >>> matrix, item_list = cost_matrix(points, euclidean_distance)
    >>> print(matrix[0, 1])  # Distance from (0,0) to (1,1)
    1.4142135623730951
    """
    n = len(items)
    arr = np.zeros((n, n))
    
    for i in range(n):
        for j in range(n):
            arr[i, j] = cost_function(items[i], items[j], **kwargs)
    
    return arr, items

def cost_matrix_to_graph(cost_matrix: np.ndarray, items: List[T]) -> nx.Graph:
    """
    Convert a cost matrix numpy array to a weighted networkx graph.
    
    Transform a symmetric cost matrix into an undirected graph where
    edge weights represent the costs between nodes. Self-loops are
    excluded from the resulting graph.

    Parameters
    ----------
    cost_matrix : numpy.ndarray
        Symmetric cost matrix with numeric values. Should be square
        with dimensions matching the length of items.
    items : List[T]
        List of items corresponding to matrix indices. Used as node
        values in the resulting graph.

    Returns
    -------
    networkx.Graph
        Undirected graph with nodes corresponding to matrix indices
        and edge weights equal to the cost matrix values. Only edges
        with positive costs are included. Node attributes 'value' 
        contain the original items.

    Examples
    --------
    Convert a simple cost matrix to a graph:
    
    >>> import numpy as np
    >>> matrix = np.array([[0, 1, 2], [1, 0, 3], [2, 3, 0]])
    >>> items = ['A', 'B', 'C']
    >>> graph = cost_matrix_to_graph(matrix, items)
    >>> list(graph.edges(data=True))
    [('A', 'B', {'weight': 1}), ('A', 'C', {'weight': 2}), ('B', 'C', {'weight': 3})]
    """
    G = nx.Graph()
    
    for i, item in enumerate(items):
        G.add_node(i, value=item)
    
    for i in range(len(items)):
        for j in range(i + 1, len(items)):
            cost = cost_matrix[i, j]
            if cost > 0:
                G.add_edge(i, j, weight=cost)
    
    return G



class Group(tuple):
    def __new__(cls, G):
        if isinstance(G, tuple):
            D = G[0]
            S = G[1]
            
            if isinstance(S, tuple):
                processed_S = []
                for item in S:
                    if isinstance(item, tuple):
                        processed_S.append(Group(item))
                    else:
                        processed_S.append(item)
                S = tuple(processed_S)
            
            G = (D, S)
        
        return super(Group, cls).__new__(cls, G)
    
    @property
    def D(self):
        return self[0]
    
    @property
    def S(self):
        return self[1]
    
    def __repr__(self) -> str:
        return super().__repr__()



from typing import Union
from fractions import Fraction
from itertools import accumulate

__all__ = [
    'cycles_to_frequency',
    'beat_duration',
    'calc_onsets',
]

def cycles_to_frequency(cycles: Union[int, float], duration: float) -> float:
    '''
    Calculate the frequency (in Hz) needed to produce a specific number of cycles within a given duration.

    Args:
    cycles (Union[int, float]): The desired number of complete cycles
    duration (float): The time duration in seconds

    Returns:
    float: The frequency in Hertz (Hz) that will produce the specified number of cycles in the given duration

    Example:
    >>> cycles_to_frequency(4, 2)  # 4 cycles in 2 seconds = 2 Hz
    2.0
    '''
    return cycles / duration

def beat_duration(ratio:Union[int, float, Fraction, str], bpm:Union[int, float], beat_ratio:Union[int, float, Fraction, str] = '1/4') -> float:
  '''
  Calculate the duration in seconds of a musical beat given a ratio and tempo.

  The beat duration is determined by the ratio of the beat to a reference beat duration (beat_ratio),
  multiplied by the tempo factor derived from the beats per minute (BPM).

  Args:
  ratio (str): The ratio of the desired beat duration to a whole note (e.g., '1/4' for a quarter note).
  bpm (float): The tempo in beats per minute.
  beat_ratio (str, optional): The reference beat duration ratio, defaults to a quarter note '1/4'.

  Returns:
  float: The beat duration in seconds.
  '''
  tempo_factor = 60 / bpm
  ratio_value  = float(Fraction(ratio))
  beat_ratio   = Fraction(beat_ratio)
  return tempo_factor * ratio_value * (beat_ratio.denominator / beat_ratio.numerator)

def calc_onsets(durations:tuple):
    return tuple(accumulate([0] + list(abs(r) for r in durations[:-1])))




# ------------------------------------------------------------------------
# Klotho/klotho/chronos/temporal_units/ut.py
# ------------------------------------------------------------------------
'''
--------------------------------------------------------------------------------------
Temporal Units
--------------------------------------------------------------------------------------
'''
from fractions import Fraction
from typing import Union, Tuple
# from ...topos.graphs.trees.algorithms import print_subdivisons
from klotho.topos.graphs.trees.algorithms import print_subdivisons
from ..rhythm_trees import Meas, RhythmTree
from ..rhythm_trees.algorithms import auto_subdiv
from klotho.chronos.utils import calc_onsets, beat_duration, seconds_to_hmsms

from enum import Enum
import pandas as pd
import networkx as nx

class ProlatioTypes(Enum):
    DURATION    = 'Duration'
    REST        = 'Rest'
    PULSE       = 'Pulse'
    SUBDIVISION = 'Subdivision'
    DURTYPES    = {'d', 'duration', 'dur'}
    RESTYPES    = {'r', 'rest', 'silence'}
    PULSTYPES   = {'p', 'pulse', 'phase'}
    SUBTYPES    = {'s', 'subdivision', 'subdivisions'}


class TemporalMeta(type):
    """Metaclass for all temporal structures."""
    pass


class Chronon(metaclass=TemporalMeta):
    __slots__ = ('_node_id', '_rt')
    
    def __init__(self, node_id:int, rt:RhythmTree):
        self._node_id = node_id
        self._rt = rt
    
    @property
    def start(self): return abs(self._rt[self._node_id]['onset_time'])
    @property
    def duration(self): return abs(self._rt[self._node_id]['duration_time'])
    @property
    def end(self): return self.start + abs(self.duration)
    @property
    def proportion(self): return self._rt[self._node_id]['proportion']
    @property
    def metric_ratio(self): return self._rt[self._node_id]['duration_ratio']
    @property
    def node_id(self): return self._node_id
    @property
    def is_rest(self): return self._rt[self._node_id]['duration_time'] < 0
    
    def __str__(self):
        return pd.DataFrame({
            'node_id': [self.node_id],
            'start': [self.start],
            'duration': [self.duration], 
            'end': [self.end],
            'is_rest': [self.is_rest],
            'proportion': [self.proportion],
            'metric_ratio': [self.metric_ratio],
        }, index=['']).__str__()
    
    def __repr__(self):
        return self.__str__()


class TemporalUnit(metaclass=TemporalMeta):
    def __init__(self,
                 span     : Union[int,float,Fraction]          = 1,
                 tempus   : Union[Meas,Fraction,int,float,str] = '4/4',
                 prolatio : Union[tuple,str]                   = 'd',
                 beat     : Union[None,Fraction,int,float,str] = None,
                 bpm      : Union[None,int,float]              = None,
                 offset   : float                              = 0
        ):
        
        self._type   = None
        
        self._rt     = self._set_rt(span, abs(Meas(tempus)), prolatio)
        
        self._beat   = Fraction(beat) if beat else Fraction(1, self._rt.meas._denominator)
        self._bpm    = bpm if bpm else 60
        self._offset = offset
        
        self._events = self._set_nodes()
    
    @classmethod
    def from_rt(cls, rt:RhythmTree, beat = None, bpm = None):
        return cls(span     = rt.span,
                   tempus   = rt.meas,
                   prolatio = rt._subdivisions,
                   beat     = beat,
                   bpm      = bpm)
    
    @property
    def span(self):
        """The number of measures that the TemporalUnit spans."""
        return self._rt.span

    @property
    def tempus(self):
        """The time signature of the TemporalUnit."""
        return self._rt.meas
    
    @property
    def prolationis(self):        
        """The S-part of a RhythmTree which describes the subdivisions of the TemporalUnit."""
        return self._rt._subdivisions
    
    # @prolationis.setter
    # def prolationis(self, prolatio: Union[tuple, str]):
    #     self._rt = self._set_rt(self.span, self.tempus, prolatio)
    #     self._events = self._set_nodes()
    
    @property
    def rt(self):
        """The RhythmTree of the TemporalUnit (returns a copy)."""
        return self._rt.copy()

    @property
    def ratios(self):
        """The ratios of a RhythmTree which describe the proportional durations of the TemporalUnit."""
        return self._rt._ratios

    @property
    def beat(self):
        """The rhythmic ratio that describes the beat of the TemporalUnit."""
        return self._beat
    
    @property
    def bpm(self):
        """The beats per minute of the TemporalUnit."""
        return self._bpm
    
    @property
    def type(self):
        """The type of the TemporalUnit."""
        return self._type
    
    @property
    def offset(self):
        """The offset (or absolute start time) in seconds of the TemporalUnit."""
        return self._offset
    
    @property
    def onsets(self):
        return tuple(self._rt.graph.nodes[n]['onset_time'] for n in self._rt.leaf_nodes)

    @property
    def durations(self):
        return tuple(self._rt.graph.nodes[n]['duration_time'] for n in self._rt.leaf_nodes)

    @property
    def duration(self):
        """The total duration (in seconds) of the TemporalUnit."""
        return beat_duration(ratio      = str(self._rt.meas * self._rt.span),
                             beat_ratio = self.beat,
                             bpm        = self.bpm
                )
    
    @property
    def time(self):
        """The absolute start and end times (in seconds) of the TemporalUnit."""
        return self._offset, self._offset + self.duration
    
    @property
    def events(self):
        return pd.DataFrame([{
            'node_id': c.node_id,
            'start': c.start,
            'duration': c.duration,
            'end': c.end,
            'is_rest': c.is_rest,
            's': c.proportion,
            'metric_ratio': c.metric_ratio,
        } for c in self._events], index=range(len(self._events)))
        
    @offset.setter
    def offset(self, offset:float):
        """Sets the offset (or absolute start time) in seconds of the TemporalUnit."""
        self._offset = offset
        self._events = self._set_nodes()
        
    def set_duration(self, target_duration: float) -> None:
        """
        Sets the tempo (bpm) to achieve a specific duration in seconds.
        
        This method calculates and sets the appropriate bpm value so that
        the TemporalUnit's total duration matches the target duration.
        
        Args:
            target_duration: The desired duration in seconds
            
        Raises:
            ValueError: If target_duration is not positive
        """
        if target_duration <= 0:
            raise ValueError("Target duration must be positive")
            
        current_duration = self.duration
        ratio = current_duration / target_duration
        new_bpm = self._bpm * ratio
        self._bpm = new_bpm
        self._events = self._set_nodes()

    def _set_rt(self, span:int, tempus:Union[Meas,Fraction,str], prolatio:Union[tuple,str]) -> RhythmTree:
        match prolatio:
            case tuple():
                self._type = ProlatioTypes.SUBDIVISION
                return RhythmTree(span = span, meas = tempus, subdivisions = prolatio)
            
            case str():
                prolatio = prolatio.lower()
                match prolatio:
                    case p if p.lower() in ProlatioTypes.PULSTYPES.value:
                        self._type = ProlatioTypes.PULSE
                        return RhythmTree(
                            span = span,
                            meas = tempus,
                            subdivisions = (1,) * tempus._numerator
                        )
                    
                    case d if d.lower() in ProlatioTypes.DURTYPES.value:
                        self._type = ProlatioTypes.DURATION
                        return RhythmTree(
                            span = span,
                            meas = tempus,
                            subdivisions = (1,)
                        )
                    
                    case r if r.lower() in ProlatioTypes.RESTYPES.value:
                        self._type = ProlatioTypes.REST
                        return RhythmTree(
                            span = span,
                            meas = tempus,
                            subdivisions = (-1,)
                        )
                    
                    case _:
                        raise ValueError(f'Invalid string: {prolatio}')
            
            case _:
                raise ValueError(f'Invalid prolatio type: {type(prolatio)}')

    def _set_nodes(self):
        """Updates node timings and returns chronon events."""
        leaf_nodes = self._rt.leaf_nodes
        leaf_durations = [beat_duration(ratio=self._rt[n]['duration_ratio'], 
                                      bpm=self.bpm, 
                                      beat_ratio=self.beat) for n in leaf_nodes]
        leaf_onsets = [onset + self._offset for onset in calc_onsets(leaf_durations)]
        
        for node, onset, duration in zip(leaf_nodes, leaf_onsets, leaf_durations):
            onset_time = -onset if duration < 0 else onset
            self._rt[node]['onset_time'] = onset_time
            self._rt[node]['duration_time'] = duration
        
        bfs_order = list(nx.bfs_tree(self._rt.graph, self._rt.root).nodes())
        non_leaf_nodes = [n for n,d in self._rt.graph.out_degree() if d > 0]
        for node in non_leaf_nodes:
            self._rt[node]['duration_time'] = beat_duration(
                ratio=str(self._rt[node]['duration_ratio']),
                bpm=self.bpm,
                beat_ratio=self.beat)
            
            current = node
            while self._rt.graph.out_degree(current) > 0:
                children = list(self._rt.graph.successors(current))
                current = min(children, key=lambda x: bfs_order.index(x))
            self._rt[node]['onset_time'] = self._rt[current]['onset_time']

        return tuple(Chronon(node_id, self._rt) for node_id in leaf_nodes)

    def __getitem__(self, idx: int) -> Chronon:
        return self._events[idx]
    
    def __iter__(self):
        return iter(self._events)
    
    def __len__(self):
        return len(self._events)
        
    def __str__(self):
        result = (
            f'Span:     {self._rt.span}\n'
            f'Tempus:   {self._rt.meas}\n'
            # f'Prolatio: {print_subdivisons(self._rt.subdivisions)}\n'
            f'Prolatio: {self._type.value}\n'
            f'Events:   {len(self)}\n'
            f'Tempo:    {self._beat} = {self._bpm}\n'
            f'Time:     {seconds_to_hmsms(self.time[0])} - {seconds_to_hmsms(self.time[1])} ({seconds_to_hmsms(self.duration)})\n'
            f'{"-" * 50}\n'
        )
        return result

    def __repr__(self):
        return self.__str__()

    def copy(self):
        """Create a deep copy of this TemporalUnit."""
        return TemporalUnit(
            span=self.span,
            tempus=self.tempus,
            prolatio=self.prolationis,
            beat=self._beat,
            bpm=self._bpm,
            offset=self._offset
        )


class TemporalUnitSequence(metaclass=TemporalMeta):
    """A sequence of TemporalUnit objects that represent consecutive temporal events."""
    
    def __init__(self, ut_seq:list[TemporalUnit]=[], offset:float=0):
        self._seq    = [ut.copy() for ut in ut_seq]
        self._offset = offset
        self._set_offsets()
    
    def _set_offsets(self):
        """Updates the offsets of all TemporalUnits based on their position in the sequence."""
        for i, ut in enumerate(self._seq):
            ut.offset = self._offset + sum(self.durations[j] for j in range(i))

    @property
    def seq(self):
        """The list of TemporalUnit objects in the sequence."""
        return self._seq

    @property
    def onsets(self):
        """A tuple of onset times (in seconds) for each TemporalUnit in the sequence."""
        return calc_onsets(self.durations)
    
    @property    
    def durations(self):
        """A tuple of durations (in seconds) for each TemporalUnit in the sequence."""
        return tuple(ut.duration for ut in self._seq)
    
    @property
    def duration(self):
        """The total duration (in seconds) of the sequence."""
        return sum(abs(d) for d in self.durations)
    
    @property
    def offset(self):
        """The offset (or absolute start time) in seconds of the sequence."""
        return self._offset
    
    @property
    def size(self):
        """The total number of events across all TemporalUnits in the sequence."""
        return sum(len(ut) for ut in self._seq)
    
    @property
    def time(self):
        """The absolute start and end times (in seconds) of the sequence."""
        return self.offset, self.offset + self.duration
        
    @offset.setter
    def offset(self, offset:float):
        """Sets the offset (or absolute start time) in seconds of the sequence."""
        self._offset = offset
        self._set_offsets()
    
    def set_duration(self, target_duration: float) -> None:
        """
        Sets the tempo (bpm) of all TemporalUnits to achieve a specific total duration in seconds.
        
        This method calculates and sets the appropriate bpm values for all TemporalUnits
        in the sequence so that the total duration matches the target duration.
        The relative durations between units are preserved by scaling all bpm values
        by the same factor.
        
        Args:
            target_duration: The desired total duration in seconds
            
        Raises:
            ValueError: If target_duration is not positive or if sequence is empty
        """
        if target_duration <= 0:
            raise ValueError("Target duration must be positive")
        
        if not self._seq:
            raise ValueError("Cannot set duration of empty sequence")
            
        current_duration = self.duration
        ratio = current_duration / target_duration
        
        for ut in self._seq:
            ut._bpm = ut.bpm * ratio
            ut._events = ut._set_nodes()
        
        self._set_offsets()
        
    def append(self, ut: TemporalUnit) -> None:
        """
        Append a TemporalUnit to the end of the sequence.
        
        Args:
            ut: The TemporalUnit to append
        """
        self._seq.append(ut.copy())
        self._set_offsets()
        
    def prepend(self, ut: TemporalUnit) -> None:
        """
        Prepend a TemporalUnit to the beginning of the sequence.
        
        Args:
            ut: The TemporalUnit to prepend
        """
        self._seq.insert(0, ut.copy())
        self._set_offsets()
        
    def insert(self, index: int, ut: TemporalUnit) -> None:
        """
        Insert a TemporalUnit at the specified index in the sequence.
        
        Args:
            index: The index at which to insert the TemporalUnit
            ut: The TemporalUnit to insert
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._seq) <= index <= len(self._seq):
            raise IndexError(f"Index {index} out of range for sequence of length {len(self._seq)}")
        
        self._seq.insert(index, ut.copy())
        self._set_offsets()
        
    def remove(self, index: int) -> None:
        """
        Remove the TemporalUnit at the specified index.
        
        Args:
            index: The index of the TemporalUnit to remove
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._seq) <= index < len(self._seq):
            raise IndexError(f"Index {index} out of range for sequence of length {len(self._seq)}")
        
        self._seq.pop(index)
        self._set_offsets()
        
    def replace(self, index: int, ut: TemporalUnit) -> None:
        """
        Replace the TemporalUnit at the specified index with a new one.
        
        Args:
            index: The index of the TemporalUnit to replace
            ut: The new TemporalUnit
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._seq) <= index < len(self._seq):
            raise IndexError(f"Index {index} out of range for sequence of length {len(self._seq)}")
        
        self._seq[index] = ut.copy()
        self._set_offsets()

    def __getitem__(self, idx: int) -> TemporalUnit:
        return self._seq[idx]
    
    def __setitem__(self, idx: int, ut: TemporalUnit) -> None:
        self._seq[idx] = ut.copy()
        self._set_offsets()

    def __iter__(self):
        return iter(self._seq)
    
    def __len__(self):
        return len(self._seq)

    def __str__(self):
        return pd.DataFrame([{
            'Tempus': ut.tempus,
            'Type': ut.type.name[0] if ut.type else '',
            'Tempo': f'{ut.beat} = {ut.bpm}',
            'Start': seconds_to_hmsms(ut.time[0]),
            'End': seconds_to_hmsms(ut.time[1]),
            'Duration': seconds_to_hmsms(ut.duration),
        } for ut in self._seq]).__str__()

    def __repr__(self):
        return self.__str__()

    def copy(self):
        """Create a deep copy of this TemporalUnitSequence."""
        copied_units = [ut.copy() for ut in self._seq]
        return TemporalUnitSequence(
            ut_seq=copied_units,
            offset=self._offset
        )


class TemporalBlock(metaclass=TemporalMeta):
    """
    A collection of parallel temporal structures that represent simultaneous temporal events.
    Each row can be a TemporalUnit, TemporalUnitSequence, or another TemporalBlock.
    """
    
    def __init__(self, rows:list[Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']]=[], axis:float = -1, offset:float=0, sort_rows:bool=True):
        """
        Initialize a TemporalBlock with rows of temporal structures.
        
        Args:
            rows: List of temporal structures (TemporalUnit, TemporalUnitSequence, or TemporalBlock)
            offset: Initial time offset in seconds
            sort_rows: Whether to sort rows by duration (longest at index 0)
        """
        self._rows = [row.copy() for row in rows] if rows else []
        self._axis = axis
        self._offset = offset
        self._sort_rows = sort_rows
        
        self._align_rows()
      
    # TODO: make free method in UT algos
    # Matrix to Block
    @classmethod
    def from_tree_mat(cls, matrix, meas_denom:int=1, subdiv:bool=False,
                      rotation_offset:int=1, beat=None, bpm=None):
        """
        Creates a TemporalBlock from a matrix of tree specifications.
        
        Args:
            matrix: Input matrix containing duration and subdivision specifications
            meas_denom: Denominator for measure fractions
            subdiv: Whether to automatically generate subdivisions
            rotation_offset: Offset for rotation calculations
            bpm: bpm in beats per minute
            beat: Beat ratio specification
        """
        tb = []
        for i, row in enumerate(matrix):
            seq = []
            for j, e in enumerate(row):
                offset = rotation_offset * i
                if subdiv:
                    D, S = e[0], auto_subdiv(e[1][::-1], offset - j - i)
                else:
                    D, S = e[0], e[1]
                seq.append(TemporalUnit(tempus   = Meas(abs(D), meas_denom),
                                        prolatio = S if D > 0 else 'r',
                                        bpm      = bpm,
                                        beat     = beat))
            tb.append(TemporalUnitSequence(seq))
        return cls(tuple(tb))

    def _align_rows(self):
        """
        Aligns the rows based on the current axis value and optionally sorts them by duration.
        If sorting is enabled, the longest duration will be at the bottom (index 0), 
        shortest at the top. If two rows have the same duration, their original order is preserved.
        """
        if not self._rows:
            return
        
        if self._sort_rows:
            self._rows = sorted(self._rows, key=lambda row: -row.duration, reverse=False)
        
        max_duration = self.duration
        
        for row in self._rows:
            if row.duration == max_duration:
                row.offset = self._offset
                continue
            
            duration_diff = max_duration - row.duration    
            adjustment = duration_diff * (self._axis + 1) / 2
            row.offset = self._offset + adjustment

    @property
    def height(self):
        """The number of rows in the block."""
        return len(self._rows)
    
    @property
    def rows(self):
        """The list of temporal structures in the block."""
        return self._rows

    @property
    def duration(self):
        """The total duration (in seconds) of the longest row in the block."""
        return max(row.duration for row in self._rows) if self._rows else 0.0

    @property
    def axis(self):
        """The temporal axis position of the block."""
        return self._axis
    
    @property
    def offset(self):
        """The offset (or absolute start time) in seconds of the block."""
        return self._offset

    @property
    def sort_rows(self):
        """Whether to sort rows by duration (longest at index 0)."""
        return self._sort_rows
    
    @sort_rows.setter
    def sort_rows(self, sort_rows:bool):
        self._sort_rows = sort_rows
        self._align_rows()
        
    @offset.setter
    def offset(self, offset):
        """Sets the offset (or absolute start time) in seconds of the block."""
        self._offset = offset
        self._align_rows()
    
    @axis.setter
    def axis(self, axis: float):
        """
        Sets the temporal axis position of the block and realigns rows.
        
        Args:
            axis: Float between -1 and 1, where:
                -1: rows start at block offset (left-aligned)
                 0: rows are centered within the block
                 1: rows end at block offset + duration (right-aligned)
                Any value in between creates a proportional alignment
        """
        if not -1 <= axis <= 1:
            raise ValueError("Axis must be between -1 and 1")
        self._axis = float(axis)
        self._align_rows()
        
    def set_duration(self, target_duration: float) -> None:
        """
        Sets the tempo (bpm) of all rows to achieve a specific total duration in seconds.
        
        This method calculates and sets the appropriate bpm values for all rows
        in the block so that the total duration matches the target duration.
        The relative durations between rows are preserved by scaling all bpm values
        by the same factor.
        
        Args:
            target_duration: The desired total duration in seconds
            
        Raises:
            ValueError: If target_duration is not positive or if block is empty
        """
        if target_duration <= 0:
            raise ValueError("Target duration must be positive")
        
        if not self._rows:
            raise ValueError("Cannot set duration of empty block")
            
        current_duration = self.duration
        ratio = current_duration / target_duration
        
        for row in self._rows:
            if hasattr(row, 'set_duration'):
                row_target = row.duration / ratio
                row.set_duration(row_target)
        
        self._align_rows()

    def prepend(self, row: Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']) -> None:
        """
        Add a temporal structure at the beginning of the block (index 0).
        
        Note: In this implementation, index 0 is considered the "bottom" row.
        
        Args:
            row: The temporal structure to add (TemporalUnit, TemporalUnitSequence, or TemporalBlock)
        """
        self._rows.insert(0, row.copy())
        self._align_rows()
        
    def append(self, row: Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']) -> None:
        """
        Add a temporal structure at the end of the block (highest index).
        
        Note: In this implementation, the highest index is considered the "top" row.
        
        Args:
            row: The temporal structure to add (TemporalUnit, TemporalUnitSequence, or TemporalBlock)
        """
        self._rows.append(row.copy())
        self._align_rows()
        
    def insert(self, index: int, row: Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']) -> None:
        """
        Insert a temporal structure at the specified index in the block.
        
        Note: Index 0 is the first row (bottom), with higher indices moving upward.
        
        Args:
            index: The index at which to insert the row
            row: The temporal structure to insert
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._rows) <= index <= len(self._rows):
            raise IndexError(f"Index {index} out of range for block of height {len(self._rows)}")
        
        self._rows.insert(index, row.copy())
        self._align_rows()

    def remove(self, index: int) -> None:
        """
        Remove the row at the specified index.
        
        Args:
            index: The index of the row to remove
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._rows) <= index < len(self._rows):
            raise IndexError(f"Index {index} out of range for block of height {len(self._rows)}")
        
        self._rows.pop(index)
        self._align_rows()
        
    def replace(self, index: int, row: Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']) -> None:
        """
        Replace the row at the specified index with a new one.
        
        Args:
            index: The index of the row to replace
            row: The new temporal structure
            
        Raises:
            IndexError: If the index is out of range
        """
        if not -len(self._rows) <= index < len(self._rows):
            raise IndexError(f"Index {index} out of range for block of height {len(self._rows)}")
        
        self._rows[index] = row.copy()
        self._align_rows()

    def __getitem__(self, idx: int) -> Union[TemporalUnit, TemporalUnitSequence, 'TemporalBlock']:
        return self._rows[idx]

    def __iter__(self):
        return iter(self._rows)
    
    def __len__(self):
        return len(self._rows)
    
    def __str__(self):
        result = (
            f'Rows:     {len(self._rows)}\n'
            f'Axis:     {self._axis}\n'
            f'Duration: {seconds_to_hmsms(self.duration)}\n'
            f'Time:     {seconds_to_hmsms(self._offset)} - {seconds_to_hmsms(self._offset + self.duration)}\n'
            f'{"-" * 50}\n'
        )
        return result

    def __repr__(self):
        return self.__str__()

    def copy(self):
        """Create a deep copy of this TemporalBlock."""
        copied_rows = [row.copy() for row in self._rows]
        return TemporalBlock(
            rows=copied_rows,
            axis=self._axis,
            offset=self._offset,
            sort_rows=self._sort_rows
        )



from enum import Enum
from klotho.utils.data_structures.dictionaries import SafeDict

class PFIELDS(Enum):
    def __call__(self):
        # return SafeDict(self.value.copy())
        return self.value.copy()

    SineEnv = SafeDict({
        'start'      : 0,
        'dur'        : 1,
        'synthName'  : 'SineEnv',
        'amplitude'  : 0.45,
        'frequency'  : 440,
        'attackTime' : 0.01,
        'releaseTime': 0.1,
        'pan'        : 0.0,
    })

    Vib = SafeDict({
        'start'      : 0,
        'dur'        : 1,
        'synthName'  : 'Vib',
        'amplitude'  : 0.45,
        'frequency'  : 440,
        'attackTime' : 0.01,
        'releaseTime': 0.1,
        'sustain'    : 0.5,
        'curve'      : 4.0,
        'pan'        : 0.0,
        'table'      : 0,
        'vibRate1'   : 3.5,
        'vibRate2'   : 5.8,
        'vibRise'    : 0.5,
        'vibDepth'   : 0.005,     
    })

    FMWT = SafeDict({
        'start'        : 0,
        'dur'          : 1,
        'synthName'    : 'FMWT',
        'frequency'    : 440,
        'amplitude'    : 0.45,
        'attackTime'   : 0.01,
        'releaseTime'  : 0.1,
        'sustain'      : 0.5,
        'idx1'         : 0.01,
        'idx2'         : 7,
        'idx3'         : 5,
        'carMul'       : 1,
        'modMul'       : 1.0007,
        'vibRate1'     : 0.01,
        'vibRate2'     : 0.5,
        'vibRise'      : 0,
        'vibDepth'     : 0,
        'pan'          : 0.0,
        'table'        : 0,
        'reverberation': 0.0,
    })

    OscTrm = SafeDict({
        'start'      : 0,
        'dur'        : 1,
        'synthName'  : 'OscTrm',
        'amplitude'  : 0.45,
        'frequency'  : 440,
        'attackTime' : 0.01,
        'releaseTime': 0.1,
        'sustain'    : 0.5,
        'curve'      : 4.0,
        'pan'        : 0.0,
        'table'      : 0,
        'trm1'       : 3.5,
        'trm2'       : 5.8,
        'trmRise'    : 0.5,
        'trmDepth'   : 0.1,
    })

    OscAM = SafeDict({
        'start'        : 0,
        'dur'          : 1,
        'synthName'    : 'OscAM',
        'amplitude'    : 0.45,
        'frequency'    : 440,
        'attackTime'   : 0.01,
        'releaseTime'  : 0.1,
        'sustain'      : 0.5,
        'pan'          : 0.0,
        'amFunc'       : 0.0,
        'am1'          : 0.75,
        'am2'          : 0.75,
        'amRise'       : 0.75,
        'amRatio'      : 0.75,
        'reverberation': 0.0,
        'visualMode'   : 0.0,
    })

    AddSyn = SafeDict({
        'start'        : 0,
        'dur'          : 1,
        'synthName'    : 'AddSyn',
        'amp'          : 0.45,
        'frequency'    : 440,
        'ampStri'      : 0.5,
        'attackStri'   : 0.1,
        'releaseStri'  : 0.1,
        'sustainStri'  : 0.8,
        'ampLow'       : 0.5,
        'attackLow'    : 0.001,
        'releaseLow'   : 0.1,
        'sustainLow'   : 0.8,
        'ampUp'        : 0.6,
        'attackUp'     : 0.01,
        'releaseUp'    : 0.075,
        'sustainUp'    : 0.9,
        'freqStri1'    : 1.0,
        'freqStri2'    : 2.001,
        'freqStri3'    : 3.0,
        'freqLow1'     : 4.009,
        'freqLow2'     : 5.002,
        'freqUp1'      : 6.0,
        'freqUp2'      : 7.0,
        'freqUp3'      : 8.0,
        'freqUp4'      : 9.0,
        'pan'          : 0.0,
        'reverberation': 0.0,
    })

    PluckedString = SafeDict({
        'start'      : 0,
        'dur'        : 1,
        'synthName'  : 'PluckedString',
        'amplitude'  : 0.45,
        'frequency'  : 440,
        'attackTime' : 0.01,
        'releaseTime': 0.1,
        'sustain'    : 0.5,
        'Pan1'       : 0.0,
        'Pan2'       : 0.0,
        'PanRise'    : 0.0,
    })
    


from typing import Callable, Union, List, Tuple, Optional
import numpy as np
from ..lattices import Lattice


class Field(Lattice):
    """
    A field is a specialized lattice that represents a function evaluated over space.
    
    Fields are n-dimensional lattices with continuous-valued functions evaluated
    at discrete lattice points. They inherit all lattice functionality while
    providing field-specific methods and semantics.
    
    Args:
        dimensionality: Number of spatial dimensions
        resolution: Number of points along each dimension
        function: Function to evaluate at each lattice point
        ranges: Spatial range for each dimension, defaults to (-1, 1) per dimension
        bipolar: If True, coordinates range from -resolution to +resolution
    """
    
    def __init__(self, 
                 dimensionality: int, 
                 resolution: Union[int, List[int]], 
                 function: Callable[[np.ndarray], np.ndarray],
                 ranges: Optional[Union[Tuple[float, float], List[Tuple[float, float]]]] = None,
                 bipolar: bool = True):
        
        super().__init__(dimensionality, resolution, bipolar)
        
        if ranges is None:
            self._ranges = [(-1.0, 1.0)] * dimensionality
        elif isinstance(ranges, tuple) and len(ranges) == 2:
            self._ranges = [ranges] * dimensionality
        else:
            if len(ranges) != dimensionality:
                raise ValueError(f"Ranges list length {len(ranges)} must match dimensionality {dimensionality}")
            self._ranges = ranges
        
        self.apply_function(function)
    
    @property
    def ranges(self) -> List[Tuple[float, float]]:
        """Spatial ranges for each dimension."""
        return self._ranges.copy()
    
    def apply_function(self, function: Callable[[np.ndarray], np.ndarray]):
        """Apply a function to all lattice points, updating their values."""
        spatial_points = self._coordinates_to_spatial_points()
        values = function(spatial_points)
        
        for coord, value in zip(self.coords(), values):
            self.set_node_attributes(coord, {'value': float(value)})
    
    def _coordinates_to_spatial_points(self) -> np.ndarray:
        """Convert integer lattice coordinates to spatial points."""
        coordinates = self.coords()
        spatial_points = []
        
        for coord in coordinates:
            spatial_point = []
            for i, c in enumerate(coord):
                if self.bipolar:
                    coord_range = 2 * self.resolution[i]
                    spatial_val = self._ranges[i][0] + (c + self.resolution[i]) * (self._ranges[i][1] - self._ranges[i][0]) / coord_range
                else:
                    coord_range = self.resolution[i]
                    spatial_val = self._ranges[i][0] + c * (self._ranges[i][1] - self._ranges[i][0]) / coord_range
                spatial_point.append(spatial_val)
            spatial_points.append(spatial_point)
        
        return np.array(spatial_points)
    
    def get_field_value(self, coordinate: Tuple[int, ...]) -> float:
        """Get the field value at a specific coordinate."""
        return super().__getitem__(coordinate)['value']
    
    def set_field_value(self, coordinate: Tuple[int, ...], value: float):
        """Set the field value at a specific coordinate."""
        self.set_node_attributes(coordinate, {'value': float(value)})
    
    def sample_field(self, points: np.ndarray) -> np.ndarray:
        """
        Sample the field at arbitrary spatial points using interpolation.
        
        Parameters
        ----------
        points : numpy.ndarray
            Array of spatial points to sample, shape (n_points, dimensionality).
            
        Returns
        -------
        numpy.ndarray
            Array of interpolated field values.
        """
        from scipy.interpolate import griddata
        
        spatial_points = self._coordinates_to_spatial_points()
        field_values = np.array([self.get_field_value(coord) for coord in self.coords()])
        
        return griddata(spatial_points, field_values, points, method='linear', fill_value=0.0)
    
    def gradient(self, coordinate: Tuple[int, ...]) -> np.ndarray:
        """
        Compute the gradient at a lattice coordinate using finite differences.
        
        Parameters
        ----------
        coordinate : tuple of int
            The lattice coordinate to compute gradient at.
            
        Returns
        -------
        numpy.ndarray
            Gradient vector at the coordinate.
        """
        gradient = np.zeros(self.dimensionality)
        neighbors = self.get_neighbors(coordinate)
        center_value = self.get_field_value(coordinate)
        
        for neighbor_coord in neighbors:
            direction = np.array(neighbor_coord) - np.array(coordinate)
            distance = np.linalg.norm(direction)
            if distance > 0:
                direction = direction / distance
                value_diff = self.get_field_value(neighbor_coord) - center_value
                gradient += direction * value_diff / distance
        
        return gradient
    
    def laplacian(self, coordinate: Tuple[int, ...]) -> float:
        """
        Compute the Laplacian at a lattice coordinate using finite differences.
        
        Parameters
        ----------
        coordinate : tuple of int
            The lattice coordinate to compute Laplacian at.
            
        Returns
        -------
        float
            Laplacian value at the coordinate.
        """
        neighbors = self.get_neighbors(coordinate)
        center_value = self.get_field_value(coordinate)
        neighbor_sum = sum(self.get_field_value(neighbor) for neighbor in neighbors)
        
        return neighbor_sum - len(neighbors) * center_value
    
    def __getitem__(self, key):
        """Allow field[coordinate] access to values for tuples, otherwise delegate to Graph."""
        if isinstance(key, tuple):
            return self.get_field_value(key)
        return super().__getitem__(key)
    
    def __setitem__(self, key, value):
        """Allow field[coordinate] = value setting for tuples."""
        if isinstance(key, tuple):
            self.set_field_value(key, value)
        else:
            raise TypeError("Only tuple keys are supported for field coordinate access")
    
    @classmethod
    def from_lattice(cls, lattice: Lattice, 
                     function: Callable[[np.ndarray], np.ndarray],
                     ranges: Optional[Union[Tuple[float, float], List[Tuple[float, float]]]] = None) -> 'Field':
        """
        Create a Field from an existing Lattice.
        
        Parameters
        ----------
        lattice : Lattice
            The lattice to convert to a field.
        function : callable
            Function to evaluate at each lattice point.
        ranges : tuple or list of tuple, optional
            Spatial ranges for the field.
            
        Returns
        -------
        Field
            A new Field instance with the same structure.
        """
        field = cls.__new__(cls)
        field._dimensionality = lattice._dimensionality
        field._resolution = lattice._resolution.copy()
        field._bipolar = lattice._bipolar
        field._coord_to_node = lattice._coord_to_node.copy()
        field._node_to_coord = lattice._node_to_coord.copy()
        field._graph = lattice._graph.copy()
        field._meta = lattice._meta.copy()
        field._next_id = lattice._next_id
        
        if ranges is None:
            field._ranges = [(-1.0, 1.0)] * field._dimensionality
        elif isinstance(ranges, tuple) and len(ranges) == 2:
            field._ranges = [ranges] * field._dimensionality
        else:
            if len(ranges) != field._dimensionality:
                raise ValueError(f"Ranges list length {len(ranges)} must match dimensionality {field._dimensionality}")
            field._ranges = ranges
        
        field.apply_function(function)
        return field 


from typing import Union
from fractions import Fraction
from itertools import cycle
from .temporal import TemporalMeta, TemporalUnit, TemporalUnitSequence, TemporalBlock, RhythmTree, Meas
from klotho.chronos.utils import beat_duration
from klotho.chronos.rhythm_trees.algorithms import segment


def segment_ut(ut: TemporalUnit, ratio: Union[Fraction, float, str]) -> TemporalUnit:
    """
    Segments a temporal unit into a new unit with the given ratio. eg, a ratio of 1/3 means
    the new unit will have a prolatio of (1, 2).
    
    Args:
    ut (TemporalUnit): The temporal unit to segment.
    ratio (Union[Fraction, float, str]): The ratio to segment the unit by.
    
    Returns:
    TemporalUnit: A new temporal unit with the given ratio.
    """
    return TemporalUnit(span=ut.span, tempus=ut.tempus, prolatio=segment(ratio), beat=ut.beat, bpm=ut.bpm)

def decompose(ut: TemporalUnit, prolatio: Union[tuple, str, None] = None, depth: Union[int, None] = None) -> TemporalUnitSequence:
    """Decomposes a temporal structure into its constituent parts based on the provided prolatio."""
    
    prolatio_cycle = []
    
    if isinstance(prolatio, tuple):
        prolatio_cycle = [prolatio]
    elif isinstance(prolatio, str) and prolatio.lower() in {'s'}:
        prolatio_cycle = [ut._rt.subdivisions]
    elif not prolatio:
        prolatio_cycle = ['d']
    else:
        prolatio_cycle = [prolatio]
        
    prolatio_cycle = cycle(prolatio_cycle)
    if depth:
        return TemporalUnitSequence([
            TemporalUnit(
                span     = 1,
                tempus   = subtree[subtree.root]['ratio'],
                prolatio = subtree.group.S if not prolatio else next(prolatio_cycle),
                beat     = ut._beat,
                bpm      = ut._bpm
            ) for subtree in [ut._rt.subtree(n) for n in ut._rt.at_depth(depth)]
        ])
    else:
        return TemporalUnitSequence([
           TemporalUnit(
               span     = 1,
               tempus   = abs(ratio),
               prolatio = next(prolatio_cycle),
               beat     = ut._beat,
               bpm      = ut._bpm
           ) for ratio in ut._rt._ratios
        ])

def transform(structure: TemporalMeta) -> TemporalMeta:
    
    match structure:
        case TemporalUnit():
            return TemporalBlock([ut for ut in decompose(structure).seq])
            
        case TemporalUnitSequence():
            return TemporalBlock([ut.copy() for ut in structure.seq])
            
        case TemporalBlock():
            raise NotImplementedError("Block transformation not yet implemented")
            
        case _:
            raise ValueError(f"Unknown temporal structure type: {type(structure)}")

def modulate_tempo(ut: TemporalUnit, beat: Union[Fraction, str, float], bpm: Union[int, float]) -> TemporalUnit:
    """
    Creates a new TemporalUnit with the specified beat and bpm, adjusting the tempus 
    to maintain the same duration as the original unit.
    
    Args:
        ut (TemporalUnit): The original temporal unit
        beat (Union[Fraction, str, float]): The new beat value
        bpm (Union[int, float]): The new beats per minute
        
    Returns:
        TemporalUnit: A new temporal unit with adjusted tempus and the specified beat/bpm
    """
    ratio = ut.duration / beat_duration(str(ut.tempus * ut.span), bpm, beat)
    new_tempus = Meas(ut.tempus * ut.span * ratio)
    
    return TemporalUnit(
        span=1,
        tempus=new_tempus,
        prolatio=ut.prolationis,
        beat=beat,
        bpm=bpm
    )

def modulate_tempus(ut: TemporalUnit, span: int, tempus: Union[Meas, Fraction, float, str]) -> TemporalUnit:
    """
    Creates a new TemporalUnit with the specified tempus, adjusting the beat/bpm 
    to maintain the same duration as the original unit.
    
    Args:
        ut (TemporalUnit): The original temporal unit
        tempus (Union[Meas, Fraction, float, str]): The new tempus value
        
    Returns:
        TemporalUnit: A new temporal unit with the specified tempus and adjusted beat/bpm
    """
    if not isinstance(tempus, Meas):
        tempus = Meas(tempus)
    
    ratio = beat_duration(str(tempus * span), ut.bpm, ut.beat) / beat_duration(str(ut.tempus * ut.span), ut.bpm, ut.beat)

    return TemporalUnit(
        span=span,
        tempus=tempus,
        prolatio=ut.prolationis,
        beat=ut.beat,
        bpm=ut.bpm * ratio
    )

def convolve(x: Union[TemporalUnit, TemporalUnitSequence], h: Union[TemporalUnit, TemporalUnitSequence], beat: Union[Fraction, str, float] = '1/4', bpm: Union[int, float] = 60) -> TemporalUnitSequence:
    beat = Fraction(beat)
    bpm = float(bpm)
    
    if isinstance(x, TemporalUnit):
        x = decompose(x)
    if isinstance(h, TemporalUnit):
        h = decompose(h)
        
    y_len = len(x) + len(h) - 1
    y = []
    for n in range(y_len):
        s = Fraction(0, 1)
        for k in range(len(x)):
            m = n - k
            if 0 <= m < len(h):
                s += modulate_tempo(x.seq[k], beat, bpm).tempus.to_fraction() * modulate_tempo(h.seq[m], beat, bpm).tempus.to_fraction()
        y.append(s)
        
    return TemporalUnitSequence([TemporalUnit(span=1, tempus=r, prolatio='d' if r > 0 else 'r', beat=beat, bpm=bpm) for r in y])
